/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef INCLUDED_CHARTABLE
#define INCLUDED_CHARTABLE

/****************************************************************
**
**	Chartable represents the shape of (some of) the characters
**	of a font.
**
**	Charpath represents the shape of one character.
**
**	There is a script that produces C++ code to initialize an
**	instance of Chartable.  See font2paths.sh.
**
**	Charpath and Chartable are designed to allow much of the
**	initialization to be done at compile time.
**
****************************************************************/

#include <string>
#include <vector>
#include <utility>
#include "point.h"
#include "path.h"

class Charpath {
public:
	int npaths;
	Path *paths;
	Point final_moveto;
	int size() const { return npaths; }
	const Path & operator[](int i) const { return paths[i]; }
};

class Chartable {
public:
	char *fontname; // name of this font
	int scale;      // the font scale at which this table is built; must be <> 0.
	                // This is the value given to the script which produced
	                // the table; bad results if 1, but 64 is OK.
	int ncharpaths; // number of characters represented
	char lowestChar; // value of the lowest character
	Charpath *cp; // pointer to array of character representations
	
	// Returns a reference to a table entry
	const Charpath & operator[](int i) const { return cp[i]; }

	// Returns the number of characters represented by this table
	int size() { return ncharpaths; }
	
	// Returns the character represented by a given index into the table
	char character(int i) { return lowestChar + i; }
	
	// returns true if the given character is in the table.
	bool contains(char c) const {
		return c >= lowestChar && c <= (lowestChar + ncharpaths - 1);
	}
	
	// returns a pointer to the entry for the given character
	const Charpath * entry(char c) const {
		return &cp[c - lowestChar];
	}
	
	// returns the value by which to multiply the table data
	// to produce the desired scale.
	double scaleFactor(double wantedScale) const {
		return  wantedScale / scale;
	}

#ifndef NO_PS_GENERATOR	
	// writes postscript code to draw the given character at
	// location (x,y) and at the given scale
	void postscript(char c,
	                double x,
	                double y,
	                double fontScale,
	                ostream &os) const;
	
	// writes postscript code to draw the given string at
	// location (x,y) and at the given scale
	void postscript(const string & text,
	                double x,
	                double y,
	                double fontScale,
	                ostream &os) const;
#endif
	
	// Adds the points representing a string to a vector.
	// Each vector<vector<Point> > represents a character.
	// Each vector<Point> represents a moveto.
	// For example, the string "i" has at least two movetos.
	// Position x,y will be the left end of the baseline.
	// The font will be scaled by wantedScaleX and wantedScaleY;
	// the scale sizes are as for the PostScript scalefont command.
	void getPoints(const string & text,
               double x,
               double y,
               double wantedScaleX,
               double wantedScaleY,
               vector<vector< vector<Point> > >& vvvec) const;	
	

};

#endif
