/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef INCLUDED_CHARTABLE2
#define INCLUDED_CHARTABLE2

/****************************************************************
**
**	Chartable2 represents the shape of (some of) the characters
**	of a font.  Member class CharEntry represents the shape of
**	one character.  Member class StringPath represents the
**	shape of a string of text.
**	
**	There is a program that produces C++ code to initialize an
**	instance of Chartable2.  See mk_table2.cpp.  Chartable2 is
**	designed to allow much of the initialization to be done at
**	compile time.  
**	
**	An instance of Chartable2 is typically linked into a
**	program.  The program invokes the Chartable2::StringPath
**	constructor, passing a pointer to the Chartable2 object, a
**	string, and other parameters.  The StringPath object
**	produces vectors of points that can be used the draw the
**	string.
**
****************************************************************/

#include <string>
#include <vector>
#include <utility>
#include "point.h"
#include "path.h"
#include "vec.h"

#define ENABLE_SHOW
#undef SHOW_NAME
#define SHOW_NAME "Chartable2"
#include "show.h"

class Chartable2 {
public:
	
	class StringPath;
	
	
	char *fontname; // name of this font
	int scale;      // the font scale at which this table is built; must be <> 0.
	                // This is the value given to the script which produced
	                // the table; bad results if 1, but 64 is OK.
	int n;          // number of characters represented
	char lowestChar; // value of the lowest character
	class CharEntry;
	CharEntry *entries; // pointer to array of character representations

	// Pathgroup contains one or more Path's which are part
	// of a character.  Two Path's are in the same PathGroup
	// if and only if one is contained within the other.
	// For example, in the letter "o", the inner circle is
	// inside the outer one, therefore the two paths would
	// be in the same PathGroup.
	class PathGroup {
	public:
		uint n;
		Path * contents;
		const uint size() const { return n; }
		const Path & operator[](uint i) const { return contents[i]; }
	};

	// CharEntry represents the table entry for one character.
	// It has an array of PathGroups, which represent
	// the fillable shapes.
	//
    class CharEntry { // must not be a subclass
    public:
    	uint n;
		PathGroup * contents; // array of n PathGroups
		Point final_moveto;
		PathGroup lines;    // may have zero size()
		uint n_separate_lines() const { return lines.size(); }
		const Path & separate_line(uint i) const {
			return lines[i];
		}
		const uint size() const { return n; }
		const PathGroup & operator[](uint i) const {
			return contents[i];
		}
	};

    // returns the number of entries
    int size() const { return n; }

    // returns the i-th entry
    const CharEntry & operator[](int i) { return entries[i]; }
		
	// returns true if the given character is in the table.
	bool contains(char c) const {
		return c >= lowestChar && c <= (lowestChar + n - 1);
	}
	
	// returns a pointer to the entry for the given character
	const CharEntry * entry(char c) const {
		return &entries[c - lowestChar];
	}
	
	// returns the value by which to multiply the table data
	// to produce the desired scale.
	double scaleFactor(double wantedScale) const {
		return  wantedScale / scale;
	}
	
	// Xform is used by the StringPath member class
	//
	struct Xform {
		double mx;
		double my;
		double tx;
		double ty;
		
		Xform(double a, double b, double c, double d)
			: mx(a), my(b), tx(c), ty(d)
			{}
		Xform() : mx(1), my(1), tx(0), ty(0) {}
		
		Point operator()(const Point & p) {
			return Point(p.x() * mx + tx, p.y() * my + ty);
		}
	};
	
	// OPath is used by the StringPath member class
	//
	class OPath : private vector<Point> {
#undef SHOW_NAME
#define SHOW_NAME "OPath"
	public:
		typedef vector<Point> super;
		
		OPath(Xform * x)
			: xfp(x) {}
		
		const Point operator[](uint i) const {
		  return (*xfp)( super::operator[](i) );
		}
		const Point untransformed(uint i) const {
		  return super::operator[](i);
		}
		void push_back(const Point & p) {
            super::push_back(p);
		}
		void reserve(uint n) { super::reserve(n); }
		uint size() const { return super::size(); }
	private:
		friend class Iter;
		Xform * xfp;
	};
	
	typedef OPath path_t;
	
};

/*
** StringPath holds the points that represent
** the shape of a string.
*/
class Chartable2::StringPath {

	private:
	typedef Util::Vec<path_t> vec_path;
	
	vec_path fillable_paths;
	vec_path line_paths;

	Xform * xforms; // one XForm per character
	
	//
	path_t & createPath(vec_path & vec, uint charnum ) {
		uint i = vec.size();
		vec.push_back( path_t( &xforms[charnum] ) );
	    return vec[ i ];
	}
	
	// Add an empty path to line_paths,
	// return a reference to it.
	path_t & createLinePath( uint i ) {
		return createPath(line_paths, i);
	}
	
	// Create an empty fillable path.
	// Return a reference to it.
	//
	path_t & createFillablePath( uint i ) {
		return createPath(fillable_paths, i);
	}
	
	public:

	// Constructs a StringPath
	// representing the shape of a string.
	//
	// The font will be scaled by wantedScaleX and wantedScaleY;
	// the scale sizes are as for the PostScript scalefont command.
	// Note: the points are in PostScript coordinates, with
	// x increasing left to right, and y increasing bottom to top.
	// To convert to a coordinate system with y increasing in
	// the opposite direction, use a negative wantedScaleY.
	//
    // The values x and y will be added to each point after scaling;
    // if both zero, the string will be positioned with the left
    // end of the baseline at (0,0).
    //
	StringPath(
		Chartable2 * ct,
        const    string & text,
        double   wantedScaleX,  // horizontal scale factor
        double   wantedScaleY,  // vertical scale factor
        bool     closePaths = true,
        double   x = 0,         // x translation
        double   y = 0          // y translation
        );

	// Destroys a StringPath
	~StringPath() { delete xforms; }

	// Returns the number of paths describing fillable polygons
	// that when combined make the shape of the string.
	uint nFillablePaths() { return fillable_paths.size(); }
	
	// Gets a reference to a path that describes one of
	// the fillable polygons of the string's shape
	const path_t & fillablePath(uint i) { return fillable_paths[i]; }
	
	// Returns the number of paths describing the
	// outline of the string
	uint nLinePaths() { return line_paths.size(); }
	
	// Gets a reference to a path that describes part of
	// the outline of the string
	const path_t & linePath(uint i) { return line_paths[i]; }

};


#endif
