#!/bin/sh

## This script creates C++ code containing
## a chartable structure (see chartable.h).
## It uses the "gs" postscript interpreter.
## $1 = font name
## $2 = font scale
## $3 = numeric code for the first character
## $4 = numeric code for the last character

font="${1:-Helvetica}"
let scale="${2:-100}"
let first="${3:-32}"
let last="${4:-126}"


# tmpfile=/tmp/"`basename $0`".$$

trace() {
	: ## echo ... "$@" ... >&2
}

err() {
	echo "$@" >&2
	exit -1
}

expect1() {
	local x
	read x
	[ "$x" == "$1" ] || err "expected '$1'"
}

expect_prefix() {
	local x
	read x
	case "$x" in
	$1*) ;;
	*) err "expected '$1'" ;;
	esac
}



#
#################################################################
#
# Create the postscript commands
#
pscmds() {

#echo /$font findfont 64 scalefont setfont
echo /$font findfont $scale scalefont setfont
echo '/s (0) def'
local i
let i="$first"
while [ "$i" -le "$last" ]
do
	echo newpath
	echo 0 0 moveto
	echo s 0 $i put 
	echo s false charpath flattenpath
	echo '(%charEnd '$i' bottomOfStack ---------------)'
	echo '{/moveto___} {/lineto} {/curveto____} {/closepath___} pathforall'
	echo '(%charStart '$i' topOfStack ----------------)'
	echo stack clear
	let i=i+1
done
echo quit

}

#################################################################
#
# Write the C++ code for one character.
#
# The path array is written to stdout.
# The data structure for the character is
# saved in the global string charstruct.
#
create_char() {

local charnum=$1
local final_moveto_x
local final_moveto_y
local npaths=0
local pathlist=""
local arrayname=$(pathArrayName $charnum)

read cmd a b c || return -1 ## end of file

[ $charnum == $a ] || err "expected char $charnum, got $a"
expect_prefix moveto
read final_moveto_y
read final_moveto_x

while create_path $charnum $npaths
do
	# note: the sequence is reversed because it's postscript
	pathlist="$pathstruct,$pathlist"
	let npaths=npaths+1
done

# Write the path array
#
echo "Path $arrayname[] = {$pathlist};"

# Add the data structure for this character
#
charstruct="{$npaths,$arrayname,$(make_point $final_moveto_x $final_moveto_y)}"

return 0
}

#################################################################
#
# Write C++ code for one point
#
make_point() {
	echo "${point_macro_name}($1,$2)"
}

#################################################################
#

pathArrayName() {
	echo q_$1
}

pointArrayName() {
	echo p_$1_$2
}

#################################################################
#
# Write C++ code for one path of a character.
#
# The point array is written to stdout.
# The path structure is saved in global variable pathstruct.
#
# Return value:
#	0 if successfully created a path
#   nonzero if reached the end of data for the character
#
# This function will abort if an error is encountered
# in the input
#
create_path() {

local pathnum=$2
local charnum=$1
local pointlist=""
local npoints=0
local arrayname=$(pointArrayName $1 $2)

read cmd a b c
case $cmd in
close*) ;;
%charEnd) return -1 ;; ## no more paths for this char
*) err "expected close for path $pathnum of char $charnum" ;;
esac

while read cmd a b c
do
	trace cmd="$cmd"
	case $cmd in
	line*|move*)
		read y
		read x
		## trailing comma is OK
		pointlist="$(make_point $x $y),$pointlist"
		let npoints=npoints+1
		if [ ${cmd:0:4} == move ]
		then
			echo "Point $arrayname[] = {$pointlist};"
			pathstruct="{$npoints, $arrayname}"
			return 0
		fi
		;;
	*)
		err "expected lineto or moveto, got $cmd"
		;;
	esac
done
err "end of file while reading path $qName"

}


#################################################################
#
# Create the C++ for characters $last to $first
#
create_code() {

local maxcharnum=$last
local mincharnum=$first
local charnum=$first
local nchars=0
local point_macro_name=P
local cplist

echo '/* Created by ' $progname '*/'
echo '#include "chartable.h"'
echo '#undef '${point_macro_name}
echo '#define '${point_macro_name}'(x,y) Point((x),(y))'

while ((charnum<=maxcharnum))
do
	trace ... charnum=$charnum ...
	create_char $charnum
	cplist="$cplist
$charstruct,"
	let nchars=nchars+1
	let charnum=charnum+1
done
echo "Charpath charpaths[]={${cplist}};"
echo "Chartable chartable={\"$font\",$scale,$nchars,$first,charpaths};"
return 0
}
#################################################################

progname=`basename $0`
pscmds | gs -dNODISPLAY -dSAFER -dQUIET -dNOPROMPT | create_code

[ -n "$tmpfile" ] && rm -f "$tmpfile"
exit 0
