/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
** This program scans the chartable,
** and produces C++ code for a Chartable2 object.
**
** This program exceeds the standard allowable
** template depth of 17.  With the GNU compiler,
** use -ftemplate-depth-N, where N is at least 20.
**
*/

#include <iostream>
#include <cstdio>
#include <string>
#include <vector>
#include <ctype.h>
#include <cstring>
#include <cassert>

#include <qcanvas.h>
#include <qapplication.h>

#include "chartable.h"
#include "bounds.h"
#include "line_intersect.h"

#undef ENABLE_SHOW
#undef SHOW_NAME
#define SHOW_NAME av0
#include "show.h"

extern Chartable chartable;

char *av0;

void errx(string s) {
	cerr << s << endl;
	exit(-1);
}

#if 0
class Int {
	int i;
public:
	Int(int a) : i(a) {}
	string str() {
		char buf[32];
		sprintf(buf,"%d",i);
		return string(buf);
	}
};
#endif

QCanvasPolygon *createPoly(QCanvas *c, const Path &p) {
	QPointArray pa(p.size());
	for (int i=0; i<p.size(); i++) {
		pa.setPoint( i, qRound(p[i].x()), qRound(p[i].y()) );
	}
	QCanvasPolygon *poly = new QCanvasPolygon(c);
	poly->setPoints(pa);
	return poly;
}

ostream & operator<<(ostream &os, const QRect &r) {
	os << "{" << r.x() << "," << r.y() << "," <<
	   r.width() << "," << r.height() << "}";
	return os;
}

ostream & operator<<(ostream &os, const QPoint &p) {
	os << "(" << p.x() << "," << p.y() << ")";
	return os;
}

ostream & operator<<(ostream &os, const QCanvasLine &l) {
	os << "{" << l.startPoint() << "," << l.endPoint() << "}";
	return os;
}


// returns true if all the line segments
// of p1 overlap the solid defined by p2
//
// the QCanvasPolygon's have to be inside
// a properly-sized QCanvas for this to work.
//
bool allSegmentsOverlap(const QCanvasPolygon *p1,
                        const QCanvasPolygon *p2)
{

	SHOW(p1->boundingRect());
    SHOW(p2->boundingRect());

	QPointArray pa = p1->areaPoints();
	bool rv = true;
	QCanvasLine * line = 0;
	uint sz = pa.size();
	for (uint i=0; i<sz; i++) {
		QPoint start = pa[i];
		QPoint end = pa[(i+1<sz) ? i+1 : 0];
		if (start == end)
			continue;
		if (!p1->canvas()) errx("! no canvas");
		line = new QCanvasLine(p1->canvas());
		line->setPoints(start.x(), start.y(), end.x(), end.y());
		SHOW2(*line, line->collidesWith(p2));
		if ( ! line->collidesWith(p2) ) {
			rv = false;
			break;
		}
	}
	delete line;
	SHOWF("allSegmentsOverlap: returning %d\n", rv);
	return rv;
}	


// Creates two QCanvasPolygons, places them on a QCanvas
// that is big enough to contain them,
// and returns the boolean result of passing them to a given function.
//
// Uses QCanvas and QCanvasPolygon
//
template <class Fn>
bool polygon_test(const Path &p1, const Path &p2, Fn fn) {

	double x,y,w,h;
	double a,b,c,d;
	vector<Point> v1; p1.toVec(v1);
	vector<Point> v2; p2.toVec(v2);
	Util::getBounds(v1, &x, &y, &w, &h);
	Util::getBounds(v2, &a, &b, &c, &d);
	
	// find the size of the canvas
	double minx = min(x,a);
	double miny = min(y,b);
	double maxx = max(x + w, a + c);
	double maxy = max(y + h, b + d);
	
	// amount to move items so they're on the canvas
	double moveX = (minx < 0) ? -minx : 0;
	double moveY = (miny < 0) ? -miny : 0;
	
	QApplication app();
	QCanvas * canvas = new QCanvas( qRound(maxx - minx),
	                        qRound(maxy - miny) );
	QCanvasPolygon * poly1 = createPoly(canvas, p1);
	poly1->moveBy(moveX, moveY);
	QCanvasPolygon * poly2 = createPoly(canvas, p2);
	poly2->moveBy(moveX, moveY);
	
	bool rv = fn( poly1, poly2 );
	delete canvas; // automatically deletes poly1, poly2
	return rv;
}


bool overlap_test(const QCanvasPolygon *a, const QCanvasPolygon *b)
{
	bool rv;
	rv = allSegmentsOverlap( a, b );
	rv = rv || allSegmentsOverlap( b, a );
	return rv;
}

// returns true if the two paths define polygons
// such that every segment of either overlaps
// the solid shape of the other.
//
// Uses QCanvas and QCanvasPolygon
//
bool overlap(const Path &p1, const Path &p2)
{
	return polygon_test(p1, p2, overlap_test);
}

bool inside_test(const QCanvasPolygon *p1,
                 const QCanvasPolygon *p2)
{
	return allSegmentsOverlap(p1, p2);
}


// returns true if 
// every segment of the first path overlaps
// the solid shape of the polygon defined by the second path.
// In this program, in the context in which this
// function is called, true means the first path is
// inside the second.
//
// Uses QCanvas and QCanvasPolygon
//
bool inside(const Path &p1, const Path &p2)
{
	return polygon_test(p1, p2, inside_test);
}



/*
 * Writer<T,C> is an abstract base for classes that
 * are vectors and create C++ code for structures
 * having this form --
 * 	{ N, ptr }
 * Where N is the number of elements in an array,
 * and ptr is a pointer to the array.
 *
 * This template, and classes derived from it,
 * cause the program to exceed the standard template depth.
 */
template<class T, char C>
class Writer : public vector<T>
{
protected:
	static int next_id;
	static int n_written;
	int id;
public:
	typedef T elem_type;
	static const char template_char = C;
	typedef Writer<T,C> writer_type;
	
	Writer() : id(next_id++) {}
	virtual ~Writer() {}
	
	// returns a string to be used as the variable name
	// of the array.
	string arrayName() const {
		char buf[32];
		sprintf(buf, "%c%d", C, id);
		string s(buf);
		return s;
	}
	
	void add(const_reference t) {
		push_back(t);
	}
	
	void add(const T *t) { add(*t); }
	
	// writeEmptyStruct writes the code for an empty data structure.
	virtual void writeEmptyStruct(ostream &os) const {
		os << "{0,0}";
	}
	
	// writeStruct writes the code for the data structure.
	// it should be called after writeAntecedents.
	virtual void writeStruct(ostream &os) const {
		if (size() > 0) {
			os << "{" << size() << "," << arrayName() << "}";
		} else {
			writeEmptyStruct(os);
		}
	}
	
	// writeInit is called once per subclass.
	// see writeAntecedents
	virtual void writeInit(ostream &os) const {}
	
	// writeMember writes one element of the array
	virtual void writeMember(const T & t, ostream &os) const = 0;
	
	// writes `<type> <name>[] = { <elem0>, <elem1>, ... };'
	virtual void writeNamedArray(ostream &os) const {
		os << memberType() << " " << arrayName() << "[]={" << endl;
		for (const_iterator i=begin(); i!=end(); ++i) {
			writeMember(*i, os);
			os << "," << endl;	
		}
		os << "};" << endl;
	}
	
	// writes all the code needed before writeStruct is called
	virtual void writeAntecedents(ostream &os) const {
		if (n_written++ == 0) {
			writeInit(os);
		}
		if (size() > 0)
			writeNamedArray(os);
	}	
	
	// returns a string to be used, in the output code,
	// as the type of the array elements.
	virtual string memberType() const = 0;

};
template<class T, char C> int Writer<T, C>::next_id = 0;
template<class T, char C> int Writer<T, C>::n_written = 0;


/* PathWriter is a vector of Points
 * that can write C++ code for a Path
 */
class PathWriter : public Writer<Point, 'p'>
{
public:
	PathWriter(const Path &pth) {
		for (int i=0; i<pth.size(); i++) {
			add(pth[i]);
		}
	}
	PathWriter() : writer_type() {}
	
	void writeInit(ostream &os) const {
		os << "typedef Point " << memberType() << ";" << endl;
	}
	
	void writeMember(const Point &p, ostream &os) const {
		os << memberType() << "(" << p.x() << "," << p.y() << ")";
	}
	
	string memberType() const { return "P"; }
	
};

// WWriter is a Writer that contains another type of Writer.
// This is an abstract class.
//
template <class T, char C>
class WWriter : public Writer<T, C>
{
public:
	
	// reimplements Writer
	virtual void writeAntecedents(ostream &os) const
	{
		// write the antecedents of the members first
		for (const_iterator i=begin(); i!=end(); ++i) {
			i->writeAntecedents(os);
		}
		// call the method of the base class
		Writer<T,C>::writeAntecedents(os);	
	}
	
	// reimplements Writer
	virtual void writeMember(const_reference r, ostream &os) const {
		r.writeStruct(os); // elements are of type Writer<...>
	}
};

// PathGroupWriter is a vector of PathWriter's.
// It writes C++ code for a PathGroup.
//
class PathGroupWriter : public WWriter<PathWriter, 'g'>
{
public:
	void writeInit(ostream &os) const {}
	
	string memberType() const { return "Path"; }
};

// CharEntryWriter is a vector of PathGroupWriter's.
// It writes C++ code for a CharEntry.
//
class CharEntryWriter : public WWriter<PathGroupWriter,'e'>
{
	const Charpath * cp;
	char ch;
	char entryChar() { return ch; }
	PathGroupWriter lines_writer; //
	void init();
	void initB();
	void create_groups_B();
	void create_lines_B();

	const PathGroupWriter & group(int i) const {
		return operator[](i);
	}
	static bool is_like_B(char c) {
		return index("B8&$",c);	
	}
	int find_nearest(const Path &q, const Point &p);
	
	typedef WWriter<elem_type, template_char> superclass;
	
public:
 	CharEntryWriter(const Charpath *x, char c)
 		: cp(x), ch(c)
 		{ is_like_B(ch) ? initB() : init(); }
	
	// reimplements Writer
	void writeInit(ostream &os) const {
		os << "typedef Chartable2::PathGroup "
		   << memberType() << ";" << endl;
	}
	
	// reimplements Writer
	string memberType() const { return "Pgrp"; }
	
	// reimplements Writer
	virtual void writeAntecedents(ostream &os) const {
		superclass::writeAntecedents(os);
		lines_writer.writeAntecedents(os);
	}	

	// reimplements Writer
	virtual void writeStruct(ostream &os) const {
		const Point &p = cp->final_moveto;
		os << "{/* ";
		if (isprint(ch))
			os << '\'' << ch << '\'';
		else
			os << hex << "0x" << (0xff & ch) << dec;
		os << " */ ";
		os << size() << ",";
		os << (size() > 0 ? arrayName() : "0") << ",";
		os << "Point(" << p.x() << "," << p.y() << "),";
		lines_writer.writeStruct(os);
		os << "}";
	}
	
};

// initialize the CharEntryWriter.
// called for most characters.
void CharEntryWriter::init() {

	PathGroupWriter *pgw = 0;
		
	// for each path of this character ...
	for (int i=0;i<cp->size(); i++) {
		SHOWF("path %d\n", i);
		const Path & path1 = (*cp)[i];
		
		// create the path writer, add to group
		PathWriter pathw;
		for (int j=0; j < path1.size(); j++) {
			pathw.add(path1[j]);
		}
		if (!pgw) pgw = new PathGroupWriter();
		pgw->add(pathw);
		
		bool end_group = true;
		
		if (i+1 < cp->size()) { // if not the last path
			const Path & path2 = (*cp)[i+1];
			bool overlap_flag = overlap(path1, path2);
			SHOW( overlap_flag );
		    if ( overlap_flag ) {
		    	SHOWF("path %d will be in same group\n", i+1);
		    	end_group = false;
		    } else {
		    	SHOWF("path %d will be in a new group\n", i+1);
		    	end_group = true;
		    }
		}
		if (end_group) {
			add(*pgw); // copy path group to character
			pgw = 0;
		}
	}
	delete pgw; // no longer needed: made a copy

} // end of CharEntryWriter::init()


// return the index of the point in Path q nearest to Point p.
//
int CharEntryWriter::find_nearest(const Path &q, const Point &p)
{
	assert(q.size() > 0);
	int rv = 0;
	double dist = q[0].distance(p);
	for (int i=1; i<q.size(); i++) {
		double d = q[i].distance(p);
		if ( dist > d ) {
			dist = d;
			rv = i;
		}
	}
	return rv;
}

// Initialize fillable groups for CharEntryWriter.
//
void CharEntryWriter::create_groups_B()
{
	int npaths = cp->size();

	// find the path which encloses all others.
	// if any path encloses some but not all, abort.
	//
	int enclosing = -1;
	vector<int> enclosed;
	for (int i=0; !enclosed.size() && i<npaths; i++) {
		const Path &path1 = (*cp)[i];
		
		// determine if path i encloses others
		for (int j=0; j<npaths; j++) {
			if (i == j) continue;
			const Path & path2 = (*cp)[j];
			bool inside_flag = inside(path2, path1);
			SHOW(inside_flag);
			if (inside_flag) {
				enclosing = i;
				enclosed.push_back(j);
			} else {
				break; // next i
			}
		}
	}
	SHOWF("char %c, enclosing path is %d\n", entryChar(), enclosing);
	
	if (enclosing == -1 || npaths > 3 || enclosed.size() != 2) {
		// this program not ready for it.
		string msg = string("unexpected layout of char '")
			+ entryChar() + '\'';
		errx(msg);
	}
	int first_enclosed = enclosed[0];
	int second_enclosed = enclosed[1];
	
	// create a PathWriter for the new outer path,
	// which will include the first enclosed path.
	PathWriter opw;
	const Path & outer = (*cp)[enclosing];
	const Path & inner = (*cp)[first_enclosed];
	int nearest = find_nearest(outer, inner[0]);
	for (int i=0; i<outer.size(); i++) {
		const Point & pt = outer[i];
		opw.push_back( pt );
		if (i == nearest) {
			for (int j=0; j<inner.size(); j++) {
				opw.add( inner[j] );
			}
			opw.add( inner[0] );
			opw.add( outer[i] );
		}
	}
	SHOW2(inner.size(), outer.size());
	SHOW(opw.size());
	
	// Create a PathGroupWriter, add to this object
	PathGroupWriter gw;
	gw.add(opw);
	PathWriter ipw( (*cp)[second_enclosed] );
	gw.add(ipw);
	add(gw);
}

// Initialize the CharEntryWriter::lines_writer.
//
void CharEntryWriter::create_lines_B() {

	// for each path of this character ...
	int cp_size = cp->size();
	for (int i=0;i<cp_size; i++) {
		SHOWF("path %d\n", i);
		const Path & path1 = (*cp)[i];

		// create the path writer
		PathWriter pathw;
		for (int ii=0; ii < path1.size(); ii++) {
			pathw.add(path1[ii]);
		}
		// add to the set of line paths
		lines_writer.add(pathw);
	}
}

// Initialize the CharEntryWriter.
// See the constructor.
//
void CharEntryWriter::initB()
{
	create_lines_B();
	create_groups_B();
}
	

/*
void CharEntryWriter::init-B() {


	PathGroupWriter *pgw = 0;
	Path * enclosing_path = 0;
	vector<Path *> enclosed;
		
	// for each path of this character ...
	uint cp_size = cp->size();
	for (int i=0;i<cp_size; i++) {
		SHOWF("path %d\n", i);
		const Path & path1 = (*cp)[i];

		for (int j=0; j<cp_size; j++) {
			if (i == j) continue;
			const Path & path2 = (*cp)[j];
			bool inside_flag = inside(path2, path1);
			SHOW(inside_flag);

			// ..........
		}
	}
}
*/

// TableWriter is a vector of CharEntryWriter's.
// It writes C++ code for a Chartable2.
//
class TableWriter : public WWriter<CharEntryWriter, 't'>
{
	Chartable *ct;
	string var_name;
	void init(); // add all entries
	bool useNameSpace() const { return false; }
	
public:
	TableWriter(Chartable *table)
		: ct(table), var_name("chartable") { init(); }
	
	string nameSpaceName() const {
		return string("Chartable2_") + ct->fontname;
	}
		
	// implements Writer
	void writeInit(ostream &os) const {
		os << "typedef Chartable2::CharEntry "
		   << memberType() << ";" << endl;
	}
	
	// implements Writer
	string memberType() const { return "CE"; }
	
	// re-implements Writer
	void writeStruct(ostream &os) const {
		os << "{";
		os << '"' << ct->fontname << "\",";
		os << ct->scale << ",";
		os << ct->ncharpaths  << ",";
		os << (unsigned int)ct->lowestChar << ",";
		os << arrayName();		
		os << "};";
	}	
	
	string name() const {
		return string("chartable2");
	};
	void setName(string nm) { var_name = nm; }
	
	void writeAll(ostream & os) const;
};

void TableWriter::init()
{
	int n = ct->size();
#if 1
	for (int i=0; i<n; i++) {
		add( CharEntryWriter( &(*ct)[i], ct->character(i) ) );
	}
#else
    //add( CharEntryWriter(ct->entry('o')) );
    add( CharEntryWriter(ct->entry('%')) );
#endif
}

void TableWriter::writeAll(ostream &os) const
{
	os << "#include \"chartable2.h\"" << endl;
	if (useNameSpace()) {
		os << "namespace " << nameSpaceName() << " {" << endl;
	}
	
	writeAntecedents(os);
	os << "Chartable2 " << name() << "=";
	writeStruct(os);
	os << ";" << endl;
	
	if (useNameSpace()) {
		os << "}; // end namespace" << endl;
    }

}

char *basename(char *name) {
	char *p = rindex(name, '/');
	return p ? p + 1 : name;
}
	

main(int argc, char **argv) {

	QApplication app( argc, argv ); // required by QCanvas

	av0 = basename( argv[0] );
	
	cout << "/* Created by " << av0 << " */\n";

	TableWriter tw(&chartable);
	tw.writeAll(cout);
}
