/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef INCLUDED_PATH
#define INCLUDED_PATH

/****************************************************************
**
**	Path represents a sequence of points in 2-dimensional space.
**	It is like a struct, with data members public, so that
**	instances can be initialized at compile time.  
**
**	The toVec() template function appends the points of a Path
**	to a vector of points.
**
****************************************************************/

#include <string>
#include <vector>
#ifdef PATH_PAIR
#include <utility>
#endif
#include "point.h"


class Path {
public:
	int npoints;
	Point *pts;
	
	// Returns the number of points in the path
	int size() const { return npoints; }
	
	// Returns the point at a given index
	const Point & operator[](int i) const { return pts[i]; }
	
	// Appends the points of this Path to a vector<Point>
	// or any similar type which holds Points and has
	// reserve() and push_back() methods
	//
	template <class Point_Vector>
	void toVec(Point_Vector & v) const {
		v.reserve(v.size() + npoints);
		for (int i=0;i<npoints;i++)
			v.push_back(pts[i]);
	}
	
	// Like the similar toVec(), this also translates
	// the points.
	//
	template <class Point_Vector>
	void toVec(Point_Vector & v,
	           double scaleX, double scaleY,
	           double translateX, double translateY) const
	{
		v.reserve(v.size() + npoints);
		for (int i=0;i<npoints;i++) {
			const Point & p = pts[i];
			double qx, qy;
			qx = p.x() * scaleX + translateX;
			qy = p.y() * scaleY + translateY;
			v.push_back( Point(qx,qy) );
		}
	}
#ifdef PATH_PAIR		
	// Returns a line segment, represented by a pair of points.
	pair<Point,Point> getPair(int i) const {
		int j = (i+1) % size();
		return make_pair(pts[i], pts[j]);
	}
#endif
};

#endif
