/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/****************************************************************
**
** Implementation of LabelDrawing
**
****************************************************************/

#include <iostream>
#include <vector>

#include "labeldrawing.h"
#include "text.h"
#include "xmark.h"
#include "polygon.h"
#include "my_canvas_polygon.h"
#include "rand.h"
#include "factory.h"

#include <qpainter.h>
#include <qpixmap.h>
#include <qcanvas.h>
#include <qlayout.h>

#undef ENABLE_SHOW
#undef SHOW_NAME
#define SHOW_NAME "LabelDrawing"
#include "show.h"

LabelDrawing::LabelDrawing( int w, int h, QWidget *parent, const char *name )
        : QWidget( parent, name ),
          origWidth(w),
          origHeight(h),
          polygon(0),
          textPolygon(0),
          moveFlag(true)
{
    npos = 16;
    //npos = 3;
    pos = 0;
    //setPalette( QPalette( QColor( 250, 250, 200) ) );

    canvas = new QCanvas();

	SHOWF("LabelDrawing(): w=%d, h=%d\n", w, h);
    canvas->resize(w, h);

	drawingView = new DrawingView(canvas, this, "drawingView");
#if 0
	drawingView->resize(w, h);
#else
    drawingView->resize(w + drawingView->verticalScrollBar()->width() + 1,
                       h + drawingView->horizontalScrollBar()->height() +1 );
#endif
	connect( drawingView, SIGNAL(released(QCanvasItem *)),
	         this,          SLOT(reposition(QCanvasItem *)) );
	
    // initialize the polygon factory
    factory.setCanvas(canvas);
	factory.setNPoints(3);
	factory.setType(Factory::Random);
	//factory.setNPoints(2);
	//factory.setType(Factory::Special);


    // Add ornaments to mark the corners of the canvas
    markCorners();

	// Draw a rectangle around the figure polygon
    markPBounds();

	// Create the figure polygon
    setPolygon();

	// Create the text polygon.
    //setText(QString("the label"));
    //setText(QString("."));
    setText(QString("B"));

	// Create the label position finder
    setLPos();

	// Move the label to the first position
    moveLabel(pos);

    // Create the configuration dialog
    configDialog = new ConfigDialog(
    	this, "Configuration Dialog", true
    	);

}

void LabelDrawing::configure()
{
    configDialog->setNPoints(factory.getNPoints());	
    configDialog->setType(factory.getType());
    configDialog->setLabelText(textPolygon->getText());
    	
	if ( configDialog->exec() == QDialog::Accepted ) {
		
		// create the new text label
		bool tflag = false;
		QString newText = configDialog->getLabelText();
		SHOW2(newText, textPolygon->getText());
		if (newText != textPolygon->getText()) {
			setText(newText, false); // don't show() it yet
			tflag = true;
		}
		factory.setNPoints(configDialog->getNPoints());
		factory.setType(configDialog->getType());
		
		// create the new polygon, and update the LPos object,
		// which depends on having the correct textPolygon.
		newPolygon();
		
		// if a new text label, display it now.
		if (tflag) {
			textPolygon->show();
			canvas->setChanged(canvas->rect());
		    canvas->update();
		}		
	}
}


// Create the label positioner
//
void LabelDrawing::setLPos()
{
	delete lpos;
#ifdef ENABLE_SHOW
	QRect r = polygon->boundingRect();
	SHOW2(r.x(), r.y());
	SHOW2(r.width(), r.height());
	SHOW2(r.x() + r.width()/2, r.y() + r.height()/2);
#endif
	lpos = new LPos(polygon->getOutline(),
	                textPolygon->getOutline(),
	                textPolygonOrigin,
	                npos);
	
}

// Create a canvas item to mark the figure
// at the point of attachment of the label.
//
void LabelDrawing::setFigureMark(int i)
{
	SHOWF("! enter LabelDrawing::setFigureMark(%i)\n", i);
	if (!lpos) throw "no lpos";
	if (!canvas) throw "no canvas";
	
	Point p = lpos->getFigurePoint(i);
	SHOWF(" figure point: "); SHOW(p);
	
	if (!figureMark)
		figureMark = new XMark(canvas, p.x(), p.y());
	else
		figureMark->move(p.x(), p.y());
	XMark *xp = figureMark;
	xp->setPen( QColor(0,0,0) );
	xp->setBrush( QColor(0,0,0) );
	xp->setZ(polygon->z()+1);
	xp->show();
	canvas->setChanged(xp->boundingRect());
    canvas->update();
	
	SHOWF("! exit LabelDrawing::setFigureMark(%i)\n", i);
}

// Create a canvas item to mark the label,
// at the point of its outline where it should
// be attached to the figure.
//
//
void LabelDrawing::setLabelMark(int i)
{
	SHOWF("! enter LabelDrawing::setLabelMark(%i)\n", i);
	if (!lpos) throw "no lpos";
	if (!canvas) throw "no canvas";
	
	Point p = lpos->getLabelOutlinePoint(i);
	//SHOW(p);
	
	if (!labelMark)
		labelMark = new XMark(canvas, p.x(), p.y());
	else
		labelMark->move(p.x(), p.y());
	XMark *xp = labelMark;
	xp->setPen( QColor(0,0,0) );
	xp->setBrush( QColor(0,0,0) );
	xp->setZ(polygon->z()+1);
	xp->show();
	canvas->setChanged(xp->boundingRect());
    canvas->update();
	
	SHOWF("! exit LabelDrawing::setLabelMark(%i)\n", i);
}

// Move the label to position i
//
void LabelDrawing::moveLabel(int i)
{
	SHOWF("! enter LabelDrawing::moveLabel\n");
	
	// mark the position on the figure
	//setFigureMark(i);
	
	if (moveFlag) {
		
		double x, y;
		lpos->getPosition(i, &x, &y);
#if 1
		SHOWF("position: "); SHOW2(x,y);
    	SHOW(lpos->getLabelOutlinePoint(i));
    	SHOW(lpos->getFigurePoint(i));
#endif
		
    	//textPolygon->hide();
    	QRect before = textPolygon->boundingRect();
    	textPolygon->move(x,y);
    	QRect after = textPolygon->boundingRect();
    	//textPolygon->show();
    	
    	canvas->setChanged(before);
    	canvas->setChanged(after);
    	canvas->update();
	
	} else {
		setLabelMark(pos);
	}
	SHOWF("! exit LabelDrawing::moveLabel\n");
	
}

// Move the text object to the nearest
// position in the list maintained by the LPos object.
//
// This is a slot to receive notification when the text
// object is moved by the DrawingView.
//
void LabelDrawing::reposition(QCanvasItem *ip)
{
	TextPolygon *t = dynamic_cast<TextPolygon *>(ip);
	if (t == textPolygon) {
		SHOWF("reposition\n");
		SHOW( polygon->collidesWith(t) );
		
		//pos = lpos->nearestFigurePos(textPolygon->getCenter());
		pos = lpos->nearestPos(textPolygon->x(), textPolygon->y());
		moveLabel(pos);
	}
}	

// Create the text object, add it to canvas.
//
void LabelDrawing::setText(const QString & s, bool showFlag = true)
{
	SHOWF("LabelDrawing::setText(%s)\n", s.latin1());
	textString = s;
	if (textPolygon) {
		drawingView->removeFromDraggable(textPolygon);
		delete textPolygon;
	}
	
	QRect tBounds(0,0,pBounds().width(),pBounds().height());
	
	//SHOWF("LabelDrawing::setText(%s): calling TextPolygon\n", s.latin1());
	SHOWF("LabelDrawing::setText: canvas=%p\n", canvas);
	textPolygon = new TextPolygon(canvas, textString, tBounds);
	textPolygonOrigin = Point(textPolygon->x(), textPolygon->y());
	//SHOWF("LabelDrawing::setText(%s): calling TextPolygon\n", s.latin1());
	
	//textPolygon->setPen( QColor(0,0,0) );
    textPolygon->setBrush( QColor(0,0,0) );
    textPolygon->setZ(256);
    if (showFlag)
    	textPolygon->show();
    	
    drawingView->setDraggable(textPolygon, true);
    SHOWF("LabelDrawing::setText(%s): exiting\n", s.latin1());
}

// Create the figure polygon, add it to canvas.
//
void LabelDrawing::setPolygon()
{
	//SHOWF(" enter LabelDrawing::setPolygon \n");
    if (polygon) {
    	//SHOWF(" deleting polygon \n");
    	delete polygon; // this automatically removes from canvas
    }
	
	//QRect bounds = pBounds();
	factory.setBounds(pBounds());
	polygon = factory.createPolygon();

    polygon->setPen( QColor(0,0,0) );
    polygon->setBrush( QBrush(QColor(32, 128, 64),  // dark green
                              Dense4Pattern) );

    polygon->setDrawnFilled(true);
    //polygon->setOutlineDrawn(true);

    polygon->setZ(256);
    polygon->show();
    //SHOWF(" exit LabelDrawing::setPolygon \n");

    //canvas->update();
}


// Return the rectangular bounds to be used
// for the figure polygon
//
QRect LabelDrawing::pBounds()
{
	int w = canvas->width()/3;
	int h = canvas->height()/3;
	//SHOW0 << "pBounds(): w=" << w << ", h=" << h << endl;
	return QRect(w, h, w, h);
}

// Mark the area of the central polygon
//
void LabelDrawing::markPBounds()
{
    QPointArray pa(5);
	QRect b = pBounds();

	//SHOW0 << "markPBounds: ";
	//SHOW0 << b.x() << "," << b.y() << "," ;
	//SHOW0 << b.width() << "," << b.height() << endl;

	pa[0] = QPoint(b.x(), b.y());
	pa[1] = QPoint(b.x(), b.y() + b.height());
	pa[2] = QPoint(b.x() + b.width(), b.y() + b.height());
	pa[3] = QPoint(b.x() + b.width(), b.y());
	pa[4] = QPoint(b.x(), b.y());
    QCanvasLine * p;
	for (int i=0; i<4; i++) {
		p = new QCanvasLine(canvas); // deleted by QCanvas::~QCanvas
		p->setPoints(pa[i].x(), pa[i].y(), pa[i+1].x(), pa[i+1].y() );   
		p->setPen( QColor(255,0,0) );
		p->setBrush( QColor(0,0,0) );
		p->move(0,0);
		p->setZ(0);
		p->show();
	}
	XMark *xp;
	for (int i=0; i<4; i++) {
		xp = new XMark(canvas, pa[i]); // deleted by QCanvas::~QCanvas
		xp->setPen( QColor(139, 69, 19) ); // chocolate4
		xp->setBrush( QColor(0,0,0) );
		xp->setZ(0);
		xp->show();
	}
}

// Add ornaments to mark the corners of the canvas
//
void LabelDrawing::markCorners()
{
	// note: the items allocated with 'new' are
	// automatically deleted when the canvas is destroyed.
	
    QPointArray pa;
    QCanvasPolygon * p = new MyCanvasPolygon( canvas );
    //SHOW0 << " p = " << p << endl;
    pa.setPoints(3, 0,0, 0,20, 20,0);
    p->setPoints(pa);
    //p->setPen( QColor(255,0,0) );
    p->setBrush( QColor(0,0,0) );
    p->move(0,0);
	p->setZ(0);
    p->show();

    p = new QCanvasPolygon( canvas );
    pa.setPoints(3, 0,0, 0,20, -20,0);
    p->setPoints(pa);
    p->setBrush( QColor(0,0,0) );
    p->move(canvas->width(),0);
	p->setZ(0);
    p->show();

    p = new QCanvasPolygon( canvas );
    pa.setPoints(3, 0,0, 0,-20, 20,0);
    p->setPoints(pa);
    p->setBrush( QColor(0,0,0) );
    p->move(0, canvas->height()-2);
	p->setZ(0);
    p->show();

    p = new QCanvasPolygon( canvas );
    pa.setPoints(3, 0,0, -20,0, 0,-20);
    p->setPoints(pa);
    p->setBrush( QColor(0,0,0) );
    p->move(canvas->width()-2, canvas->height()-2);
	p->setZ(0);
    p->show();
}

QSize LabelDrawing::sizeHint() const
{
	return QSize(origWidth, origHeight);
}

LabelDrawing::~LabelDrawing()
{
	//SHOW0 << "enter LabelDrawing::~LabelDrawing \n";
	//delete polygon;
	//SHOW0 << " deleting canvas...\n";
	delete canvas;
	//SHOW0 << " done\n";
	delete drawingView;
	delete lpos;
	delete configDialog;
	
	// note: the canvas items are deleted automatically
	// by the canvas destructor
	
	SHOW0 << "exit LabelDrawing::~LabelDrawing \n";
	
}

void LabelDrawing::next( int step=1 )
{
	SHOW0 << "enter LabelDrawing::next \n";
	
	//SHOW(canvas->allItems().size());
	//textPolygon->move(textPolygon->x()+10,textPolygon->y()+10);
	
	pos = (pos + step) % npos;
	if (pos < 0) pos += npos;
	moveLabel(pos);
	////////////////
	
	SHOW0 << "exit LabelDrawing::next \n";

}

void LabelDrawing::prev( int step=1 )
{
	next(- step);
}

void LabelDrawing::newPolygon(  )
{
#ifdef ENABLE_SHOW
	SHOW0 << " LabelDrawing::newPolygon \n";
	SHOW(canvas->allItems().size());

	QSize vsize = drawingView->size();
	QSize csize = canvas->size();
    if (&vsize || &csize) ;
    SHOW2(vsize.width(), vsize.height());
	SHOW2(csize.width(), csize.height());
#endif
	
	setPolygon();
	setLPos();
	moveLabel(pos = 0);
	
	canvas->setChanged(canvas->rect());
    canvas->update();
	
}


//QSizePolicy LabelDrawing::sizePolicy() const
//{
//    return QSizePolicy( QSizePolicy::Expanding, QSizePolicy::Expanding );
//}

