/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef LabelDrawing_H
#define LabelDrawing_H

/****************************************************************
**
** Declaration of LabelDrawing
**
** LabelDrawing is the heart of the program.
** It is a QWidget that displays the drawing,
** and provides two ways to change it.
** It has Qt slots that allow the drawing
** to be modified, and uses a DrawingView,
** which allows the label to be dragged to
** a new location.
**
****************************************************************/

#include <qwidget.h>
#include <qcanvas.h>
#include <qptrlist.h>

#include "xmark.h"
#include "lpos.h"
#include "text.h"
#include "polygon.h"
#include "factory.h"
#include "config_dialog.h"
#include "drawing_view.h"

class LabelDrawing : public QWidget
{
    Q_OBJECT
public:

	/*
	 * Creates a LabelDrawing having width w, height h, and 
	 * the given parent widget and QObject name.
	 */
    LabelDrawing( int w, int h, QWidget *parent=0, const char *name=0 );
    ~LabelDrawing();

	// Reimplements QWidget::sizeHint
    QSize sizeHint() const;


public slots:

	/*
	 * Moves the label to the next position
	 */
    void next(int step=1);

	/*
	 * Moves the label to the previous position
	 */
    void prev(int step=1);

	/*
	 * Creates a new figure polygon, and places the
	 * label at the first position.
	 */
    void newPolygon();

	/*
	 * Creates a new label, using the given string.
	 */
    void setText(const QString & s, bool showFlag = true);

	/*
	 * Invokes a configuration dialog to change the drawing.
	 */
    void configure();

protected:
    // void paintEvent( QPaintEvent * );

private slots:
	void reposition(QCanvasItem *ip);
	
private:

	// don't allow copy constructor or assignment
	// This makes it easier to manage pointers
	// that the destructor deletes.
	//
	LabelDrawing(const LabelDrawing &) : QWidget()
		{ throw "LabelDrawing: no copy constructor\n"; }
	LabelDrawing& operator=(const LabelDrawing &)
		{ throw "LabelDrawing: no assignment operator\n"; }

    int origWidth;
    int origHeight;
    int npos; // number of label positions
    int pos;  // current position, 0 to npos-1
    Polygon * polygon;
    QString textString; // text to be drawn
    TextPolygon * textPolygon;
    Point textPolygonOrigin;
    QCanvas * canvas;
    DrawingView * drawingView;
    XMark * figureMark;
    XMark * labelMark;
    bool moveFlag; // true means the label can be moved
    LPos * lpos;
    Factory factory;
    ConfigDialog * configDialog;

    void markCorners();
    void markPBounds();
    QRect pBounds();
	void setPolygon();
	void setFigureMark(int i);
	void setLabelMark(int i);
	void moveLabel(int i);
	void setLPos();
};


#endif // LabelDrawing_H
