/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 *	
 *	
 */
#include <cassert>

#include "lpos.h"
#include "outline.h"
#include "line_intersect.h"

#undef ENABLE_SHOW
#undef SHOW_NAME
#define SHOW_NAME "LPos"
#include "show.h"

bool LPos::FindAngle::operator()(const Outline::PtInfo *pa) {
	double a = pa->getRadians();
	double b = pa->nextInfo()->getRadians();
	//SHOW2(a,b);
	a = Outline::canonicalRadians(a); // redundant?
	b = Outline::canonicalRadians(b);
	//SHOW2(a,b);
	
	//double r = Outline::canonicalRadians(radians);
	found = false;
	double &r = radians;
	//SHOW(r);
	if (a >= b) {
		found = ( a >= r && r >= b );
	} else {
		found = ( a >= r || r >= b );
	}
	
	return found;
}

bool LPos::FindMaxAtAngle::operator()(const Outline::PtInfo *pa) {
	//SHOW0 << "LPos::FindMaxAtAngle::operator() --\n";
	Point p = pa->getPoint();
	double v = angleVal(p, radians);
	//SHOW2(p,v); SHOW2(maxPoints.size(), maxVal);
	if (maxPoints.empty()) {
		maxVal = v;
		maxPoints.push_back(p);
	} else if (v > maxVal) {
		maxVal = v;
		maxPoints.clear();
		maxPoints.push_back(p);
	} else if (v == maxVal && p != maxPoints[0]) {
		maxPoints.push_back(p);
	}
	//SHOW0 << "--\n";
	return false;
}

/*
 * Returns how far a point is in a given direction.
 *
 * Imagine a line from the origin at the given angle,
 * intersecting a perpendicular line thru the
 * given point.  The returned value is the distance
 * from the origin to that perpendicular line.
 * It might be a negative number.
 *
 */
double LPos::FindMaxAtAngle::angleVal(Point p, double radians) {
	// rotate the point and get the new x value
	double rad = -radians;
	double cosa = cos(rad);
	double sina = sin(rad);
	double newx = cosa * p.x() + -sina * p.y();
	//double newy = cosa * y + sina * x;
	//SHOW2(newx, newy);
	return newx;
}

// initialize the LPos object by setting its
// first table entry to contain the point of "attachment"
// on the figure of the first label position.
//
void LPos::init()
{
	int oct = degreesToOctant(startDegrees);
	const Outline::PtInfo * ip;
	FindAngle f(startRadians);
	ip = figureOutline->findPoint(oct, f);
	if (!ip)
		ip = figureOutline->findPoint(prevOct(oct), f);
	if (!ip)
		ip = figureOutline->findPoint(nextOct(oct), f);
	if (!ip)
		ip = figureOutline->findPoint(prevOct(prevOct(oct)), f);
	
	if (!ip) {
		SHOW0 << "---------\n";
		SHOW(startRadians);
		SHOW(figureOutline->size());
		vector<Point> v;
		figureOutline->getPoints(v);
		for (unsigned int i=0; i<v.size(); i++) {
			SHOW(v[i]);
		}
		SHOW0 << "---------\n";
		throw "LPos::init: angle not found"; // should not happen
	}
	
	// now ip points to the start of the line segment.
	// find the intersection
	Point a = ip->getPoint();
	Point b = ip->nextInfo()->getPoint();
	SHOW2(a, b);
	SHOW(figureOutline->getCenter());
	Point p = Util::findIntersection(figureOutline->getCenter(),
	                           startRadians,
	                           a,
	                           b);
	
	table[0].figurePoint = p;
	table[0].segmentPtInfo = ip;
	table[0].segmentDistance = a.distance(p);
	nFigurePosSet = 1;
}

int LPos::degreesToOctant(double d)
{
	d = Outline::canonicalDegrees(d); // 0 to 359.999...
	if (d==0) return Outline::octMAXX;
	else if (d < 90) return Outline::octHIHI;
	else if (d == 90) return Outline::octMAXY;
	else if (d < 180) return Outline::octLOHI;
	else if (d == 180) return Outline::octMINX;
	else if (d < 270) return Outline::octLOLO;
	else if (d == 270) return Outline::octMINY;
	else if (d < 360) return Outline::octHILO;
	else return Outline::octNONE; // should never happen
}

int LPos::radiansToOctant(double r) {
	return degreesToOctant( r * 180.0 / M_PI );
}

inline bool LPos::isCornerOct(int oct) {
	return Outline::isCornerOct(static_cast<Outline::OctNumber>(oct));
}


// Set the translation amounts in an entry of the position table.
//
void LPos::compute_translation(int i)
{
	Pos *t = &table[i];
	
	// Find the "angle of attachment", which is
	// an angle from the point of attachment (on the outline
	// of the figure) toward the inside of the figure outline.
	// If the point is on an apex, the angle goes toward
	// the center of the figure outline's rectangular bounds.
	// Otherwise, the angle is perpendicular to the segment
	// of the outline which contains the point.
	Point fp = getFigurePoint(i); // side effects
	const Outline::PtInfo *seg_start = t->segmentPtInfo;
	const Outline::PtInfo *seg_end   = seg_start->nextInfo();
	double angle;
	Point p1 = seg_start->getPoint();
	Point p2 = seg_end->getPoint();
	bool atFigureApex = true;
	if (fp == p1)  angle = seg_start->getRadians();
	else if (fp == p2)  angle = seg_end->getRadians();
	else {
		atFigureApex = false;
		double seg_angle = atan2( p2.y() - p1.y(), p2.x() - p1.x() );
		// Add 1/2 PI or 90 degrees.
		// This assumes the outline traversal order is positive-y
		// to positive-x (clockwise if the coordinate system is viewed
		// as having y increasing upward x increasing to the right).
		angle = seg_angle + M_PI_2;
	}
	angle = Outline::canonicalRadians(angle + M_PI); // add 180 degrees
	//
	// Now we have the angle.
	// Get the points of the label outline which are
	// farthest in that direction.
	vector<int> voct(3);
	int oct = radiansToOctant(angle);
	voct.push_back(oct);
	if (isCornerOct(oct)) {
		voct.push_back(prevOct(oct));
		voct.push_back(nextOct(oct));
	}
	FindMaxAtAngle fmaa(angle);
	for (unsigned int i=0; i<voct.size(); i++) {
		labelOutline->findPoint(voct[i], fmaa);
	}
	vector<Point> lpoints(fmaa.getMaxPoints());
	//SHOW2(atFigureApex, lpoints.size());
	if (lpoints.size() > 2) {
		//throw "LPos: too many points";
		cerr << "LPos: too many points" << endl;
	}
	Point lp;
	if (lpoints.size()>1) {
		t->surfaces = !atFigureApex;
		//SHOW2(lpoints[0], lpoints[1]);
		lp = lpoints[0].midPoint(lpoints[1]);
	} else {
		t->surfaces = false;
		lp = lpoints[0];
	}
		
	t->labelPoint = lp;
	t->trans_x = fp.x() - lp.x();
	t->trans_y = fp.y() - lp.y();
	//SHOW(lp);
	//SHOW2(t->trans_x, t->trans_y);
	t->has_trans = true;
}

/*
 * Gets the x and y coordinates to which the label should be
 * moved for position i.
 *
 * The coordinates are relative to the label origin, which
 * was given to the LPos constructor, and which
 * represents the location of labelOutline.
 * For example, if the label origin is (2,3). and the label
 * should be moved from the location of labelOutline by (+1,+2),
 * then this method will output *x=3 and *y=4.
 */
void LPos::getPosition(int i, double *x, double *y)
{
	SHOW0 << "enter LPos::getPosition, i=" << i << endl;
	if (i<0 || i>npos) throw "LPos::getPosition: bad position number";
	*x = 0;
	*y = 0;
	Pos *t = &table[i];
	if (!t->has_trans) {
		compute_translation(i);
	}
	*x = labelOrigin.x() + t->trans_x;
	*y = labelOrigin.y() + t->trans_y;
    SHOW0 << "leave LPos::getPosition, i=" << i << endl;
}

Point LPos::getPosition(int i) {
	double x, y;
	getPosition(i, &x, &y);
	return Point(x,y);
}
	

Point LPos::getLabelOutlinePoint(int i) {
	double x, y;
	getPosition(i, &x,&y);
	return table[i].labelPoint;
}

void LPos::figureStep()
{
	assert(nFigurePosSet < npos);

	Pos * u = &table[nFigurePosSet]; // first unused entry
	u->segmentPtInfo = (u-1)->segmentPtInfo;
	u->segmentDistance = (u-1)->segmentDistance;
	double segLen = (u-1)->segmentPtInfo->getDist();
	double stepRemainder = figureOutline->getTotalLength() / npos;
	while (stepRemainder > 0) {
		//SHOW(stepRemainder);
		if (segLen - u->segmentDistance >= stepRemainder) {
				
			// This step is completed on the current segment.
			// Add an entry to the figure position table.
			// The figure point might be at the end of
			// the segment (stepRemainder equals the trailing
			// part of the segment), but not at the beginning.
			u->segmentDistance += stepRemainder;
			double ratio = u->segmentDistance / segLen;
#if 0
			double px, py, qx, qy;
			u->segmentPtInfo->getPoint(&px, &py);
			u->segmentPtInfo->nextInfo()->getPoint(&qx, &qy);
			//SHOW2(px,py); SHOW2(qx,qy);
			u->figurePoint = Point( px + (qx - px) * ratio,
									py + (qy - py) * ratio );
#else
			Point p = u->segmentPtInfo->getPoint();
			Point q = u->segmentPtInfo->nextInfo()->getPoint();
			u->figurePoint = p.toward(q, ratio);
			//SHOW2(p,q);
#endif
			//SHOW(u->figurePoint);
			nFigurePosSet++;
			stepRemainder = 0; // break loop
				
		} else {
			stepRemainder -= (segLen - u->segmentDistance);
			// make u point to the start of the next segment
			u->segmentPtInfo = u->segmentPtInfo->nextInfo();
			u->segmentDistance = 0;
			segLen = u->segmentPtInfo->getDist();
		}	
	}
}

/*
 * Gets the point of attachment for a given position
 *
 * Assumes that nFigurePosSet is >= 0.
 *
 * This function can change nFigurePosSet.
 */
Point LPos::getFigurePoint(int i)
{
	//SHOW0 << "LPos::getFigurePoint i=" << i << endl;
	//SHOW2(nFigurePosSet, npos);
	
	if (i<0 || i>= npos)
		throw("LPos::getFigurePoint: bad parameter");
	
	// must have at least one figure point
	if (nFigurePosSet <= 0)
			throw("LPos::getFigurePoint: no table entries set\n");
	
	// Find all figure points before this one
	//
	while (nFigurePosSet < i + 1) {
		figureStep();
	}
	
	//SHOW0 << "LPos::getFigurePoint nFigureSet=" << nFigurePosSet << endl;	
	return table[i].figurePoint;
}

int LPos::nearestFigurePos(const Point & p, Point *out)
{
	int minPos = 0;
	Point minPt = getFigurePoint(0);
	double minDist = p.distance(minPt);
	for (int i=1; i < npos; ++i) {
		Point q = getFigurePoint(i);
		double dist = p.distance(q);
		if (dist < minDist) {
			minPt = q;
			minPos = i;
			minDist = dist;
		}
	}
	if (out) *out = minPt;
	return minPos;
}

int LPos::nearestPos(double x, double y, Point *out)
{
	Point p(x,y);
	int minPos = 0;
	Point minPt = getPosition(0);
	double minDist = p.distance(minPt);
	for (int i=1; i < npos; ++i) {
		Point q = getPosition(i);
		double dist = p.distance(q);
		if (dist < minDist) {
			minPt = q;
			minPos = i;
			minDist = dist;
		}
	}
	if (out) *out = minPt;
	return minPos;
}
