/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef INCLUDED_LPOS
#define INCLUDED_LPOS

/****************************************************************
**
** LPos is a "label positioner", that produces positions
** for a shape called the "label" so that it is tangent
** to another shape called the "figure".
**
****************************************************************/

#include <functional>
#include <cmath>

#include "point.h"
#include "outline.h"

#undef ENABLE_SHOW
#undef SHOW_NAME
#define SHOW_NAME "LPos"
#include "show.h"

class LPos {
	Outline * figureOutline;
	Outline * labelOutline;
	Point labelOrigin; // origin of label when its outline created
	int npos;
	double startDegrees;
	double startRadians;
	//Point figureCenter;
	void figureStep();
	
	struct Pos {
		double trans_x; // the translation of the label
		double trans_y;
		Point labelPoint; // for debug; point of label to move to figurePoint
		bool has_trans;  // true when calculations complete for this position
		bool surfaces; // facets of the outlines meet at this position
		Point figurePoint;  // point at which the label will be "attached"
		const Outline::PtInfo * segmentPtInfo; // start of the segment on figure
		double segmentDistance; // distance of figurePoint from start of segment
		Pos() : has_trans(false), segmentPtInfo(0) {}
	};
	Pos * table;
	//int nSet;
	int nFigurePosSet; // number of entries with figurePoint set
	
	// FindAngle is a function object, used with
	// Outline::findPoint(), for finding a point of the figure
	// outline that is the start of the segment which intersects a line
	// from the center of the figure.
	// The constructor is given the angle to be found.
	// The operator()() method is given a pointer to
	// the Outline::PtInfo object for the start of a segment,
	// and returns true if a line at the given angle crosses the segment.
	//
	class FindAngle : public unary_function<const Outline::PtInfo *, bool> {
		double radians;
		bool found;
	public:
		FindAngle(double r) : radians(r), found(false) {
			radians = Outline::canonicalRadians(radians);
		}
		
		bool operator()(const Outline::PtInfo *pa);
		bool isFound() { return found; }
	};
	
	// FindMaxAtAngle is a function object, used with
	// Outline::findPoint(), for finding a point of the figure
	// outline that is farthest in a particular direction.
	// The constructor is given the direction as an angle in radians.
	// The operator()() method is given a pointer to
	// the Outline::PtInfo object for the start of a segment.
	//
	class FindMaxAtAngle : public unary_function<const Outline::PtInfo *, bool> {
		double radians;
		vector<Point> maxPoints;
		double maxVal;
		Outline::PtInfo *ptiMax;
		static double angleVal(Point p, double radians);
	
	public:
		FindMaxAtAngle(double r) : radians(r) {
			radians = Outline::canonicalRadians(radians);
		}
		
		bool operator()(const Outline::PtInfo *pa);
		vector<Point> getMaxPoints() const { return maxPoints; }
	};
	
	void init();
	static double deg2rad(double x) { return x * M_PI / 180.0; }
	static int degreesToOctant(double d);
	static int radiansToOctant(double d);
	static inline bool isCornerOct(int);
	int prevOct(int i) { return (i+7)%8; }
	int nextOct(int i) { return (i+1)%8; }
	inline void compute_translation(int i);
	
	//
	// Don't allow copy by constructor or assignment.
	// This simplifies management of pointers deleted by the destructor.
	//
	LPos(const LPos &) { throw "LPos: no copy constructor\n"; }
	LPos& operator=(const LPos &) { throw "LPos: no assignment operator\n"; }
	
public:

	LPos(Outline * fo, // outline of the figure
	     Outline * lo, // outline of the label
	     Point l_origin = Point(0,0), // label origin when *lo built
	     int n = 20,   // number of label positions
	     double a = 0  // starting angle from center of figure
		 )
		 : figureOutline(fo),
		   labelOutline(lo),
		   labelOrigin(l_origin),
		   npos(n),
		   startDegrees(Outline::canonicalDegrees(a))
	{
		if (!fo | !lo | n<1) throw ("LPos::LPos: bad parameter");
		table = new Pos[npos];
		startRadians = Outline::canonicalRadians( deg2rad(startDegrees) );
		//figureCenter = figureOutline->getCenter();
		init();
	}
	
	~LPos() {
		//cerr << "enter ~LPos\n";
		delete [] table;
		//cerr << "exit ~LPos\n";
	}
		
	//void operator()(Point p, int type);
	
	/*
	** Gets the (x,y) location of the label
	** when it is in position i
	*/
	void getPosition(int i, double *x, double *y);
	
	/*
	** Gets the (x,y) location of the label
	** when it is in position i.
	** This is like the above method.
	*/
	Point getPosition(int i);
	
	/*
	** Gets the point on label that touches the placement
	** point on the figure, when the label is in position i.
	*/
	Point getLabelOutlinePoint(int i);
	
	/*
	** Gets the placement point on figure for position i.
	*/
	Point getFigurePoint(int i);
	
	/*
	** Of the placement points on the figure,
	** find the one nearest the given Point p.
	** Returns the position number.
	** If `out' is non-zero, *out is set to the location on the figure.
	*/
	int nearestFigurePos(const Point & p, Point *out = 0);
	
	/*
	** Find the label position nearest the given (x,y) location.
	** Returns the position number.
	** If `out' is non-zero, *out is set to the label position
	*/
	int nearestPos(double x, double y, Point *out = 0);
};

#endif
