/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <functional>
#include <cmath>
#include "outline2.h"
#include "outline_callback.h"

#undef ENABLE_SHOW
#undef SHOW_NAME
#define SHOW_NAME "Outline"
#include "show.h"

Outline::PtInfo::PtInfo(Outline *m, int i, int t)
	: mp(m), index(i), type(t)
{
	invalidateMeasurements();
}


Point Outline::PtInfo::getPoint() const
{
	return mp->getPoint(index);
}

double Outline::PtInfo::getDist() const
{
	if (!distValid) {
		Point a = getPoint();
		Point b = nextInfo()->getPoint();
		//dist = hypot(b.x()-a.x(), b.y()-a.y());
		dist = a.distance(b);
		distValid = true;
	}
	return dist;
}

double Outline::PtInfo::getRadians() const
{
	if (!angleValid) {
		Point c = mp->getCenter();
		Point p = getPoint();
		double x = p.x()-c.x();
		double y = p.y()-c.y();
		angle = atan2(y,x);
		angleValid = true;
	}
	return angle;
}

const Outline::PtInfo * Outline::PtInfo::nextInfo() const
{
	return mp->infoPtr( (index + 1) % mp->info.size() );
}

const Outline::PtInfo * Outline::PtInfo::prevInfo() const
{
	int tableSize = static_cast<int>(mp->info.size());
	return mp->infoPtr( (index + tableSize - 1) % tableSize );
}

ostream& operator<<(ostream& os, const Outline::PtInfo& p) {
	int type = p.getType();
	os << "{PtInfo: \n";
	os << " mp=" << p.mp << endl;
	os << " index=" << p.getIndex() << endl;
	//os << " point=" << p.getPoint() << endl;
	os << " type=" << type <<
	      "(0x" << hex << type << ") " << dec;
	Outline::sendType(os, type) << endl;
	os << " radians=" << p.getRadians()
	   << " (" << p.getRadians() *180.0 / M_PI << " deg)"  << endl;
	os << " dist=" << p.getDist() << endl;
	os << "}";
	return os;
/*
Point getPoint() const;
int getIndex() { return index; }
int getType() const { return type; }
double getRadians() const { return angle; }
double getDist() const { return dist; }
*/
}

ostream& Outline::sendType(ostream& os, int t) {
	os << '.';
    if (t & OutlineConstants::MINX) { os << "MINX."; }
    if (t & OutlineConstants::LOHI) { os << "LOHI."; }
    if (t & OutlineConstants::MAXY) { os << "MAXY."; }
    if (t & OutlineConstants::HIHI) { os << "HIHI."; }
    if (t & OutlineConstants::MAXX) { os << "MAXX."; }
    if (t & OutlineConstants::HILO) { os << "HILO."; }
    if (t & OutlineConstants::MINY) { os << "MINY."; }
    if (t & OutlineConstants::LOLO) { os << "LOLO."; }
 	return os;
}


Outline::Outline()
{
	x = 0;
	y = 0;
	w = 0;
	h = 0;
	
	translateX = 0;
	translateY = 0;
	scaleX = 1;
	scaleY = 1;
	
	//npoints = 0;

}

Outline::Outline(const vector<Point> &vin)
{
	init(vin.begin(), vin.end());
}

template <class In>
Outline::Outline(In b, In e)
{
	init(b,e);
}

//void Outline::init(const vector<Point> &vin) {
template <class In>
void Outline::init(In b, In e) {

	Outline2 m(b,e);
    x = m.getMinx();
	y = m.getMiny();
	w = m.getMaxx() - x;
	h = m.getMaxy() - y;
	translateX = 0;
	translateY = 0;
	scaleX = 1;
	scaleY = 1;
	
	for (int i=0; i<8; i++)
		setOctStartIndex(i, -1);

    // Fill in the vector of points
    Point center(getCenter());
	OutlineCallback c(this, /* &points, &info, */ center);
	m.traverseWithCallback(c);
	c.finish();
	totalLength = -1;
#if 0	
	SHOW(info.size());
	SHOW0 << "::::::::\n";
	for (unsigned int i=0; i<info.size(); i++) {
		cerr << info[i] << endl;
		cerr << points[i] << endl;	
	}
	SHOW0 << "::::::::\n";
#endif	
	
	
}

Outline::~Outline()
{
	
}

const vector<Point> &Outline::getVector() const {
	return points;
}

Point Outline::getPoint(int i) const {
	return transformPoint(points[i].x(), points[i].y());
}

void  Outline::getPoint(int i, double * ox, double * oy) const {
	transformPoint(points[i].x(), points[i].y(), ox, oy);
}

void Outline::getPoints(vector<Point> &vout) const
{
	//cerr << "enter getPoints\n";
	//SHOW(points.size());
	//SHOW(types.size());
	//SHOW(points[0]);
	//SHOW(points[points.size()-1]);
	for (unsigned int i=0; i<points.size(); i++)
	{
		//cerr << i << endl;
		vout.push_back( getPoint(i) );
	}
	//cerr << "exit getPoints\n";
}

/*
 * .... Note: if used, move these definitions
 * of template methods to the .h file.

template <class Callback>
void Outline::getPointsByCallback(Callback *callback) const
{
	for (unsigned int i=0; i<points.size(); i++)
		(*callback)(getPoint(i), types[i]);
}

template <class Callback>
void Outline::getPointsByCallback(Callback &callback) const
{
	getPoints(&callback);
}
*/



#if 0
// convert degrees to number >= 0 and < 360
double Outline::canonicalDegrees(double x) {
	int xIntDividend = int(x / 360);
	double val = x - (xIntDividend * 360);
	return val < 0 ? 360 + val : val;
}
#endif
inline double Outline::canonicalNumber(double x, double lim) {
	int xIntDividend = int(x / lim);
	double val = x - (xIntDividend * lim);
	return val < 0 ? lim + val : val;
}

// convert degrees to number >= 0 and < 360
double Outline::canonicalDegrees(double x) {
	return canonicalNumber(x, 360.0);
}

// convert radians to number >= 0 and < (2 * pi)
double Outline::canonicalRadians(double x) {
	return canonicalNumber(x, 2 * M_PI);
}
	
void Outline::setOctStartIndex(int oct, int index) {
	//cerr << "setOctStartIndex: "; SHOW2(oct, index);
	octStart[oct] = index;
}

void Outline::setOctEndIndex  (int oct, int index) {
	//cerr << "setOctEndIndex: "; SHOW2(oct, index);
	octEnd[oct] = index;
}


#if 0	
Outline::OctNumber Outline::degreesToOctant(double d)
{
	d = canonicalDegrees(d); // 0 to 359.999...
	if (d==0) return octMAXX;
	else if (d < 90) return octHIHI;
	else if (d == 90) return octMAXY;
	else if (d < 180) return octLOHI;
	else if (d == 180) return octMINX;
	else if (d < 270) return octLOLO;
	else if (d == 270) return octMINY;
	else if (d < 360) return octHILO;
	else return octNONE; // should never happen
}
#endif


Point Outline::getCenter() const
{

	return transformPoint(x + w/2, y + h/2);

}

void Outline::getCenter(double *ox, double *oy) const
{
	transformPoint(x + w/2, y + h/2, ox, oy);
}


void Outline::getVectorBounds(double *xp, double *yp, double *wp, double *hp) const
{
	if (xp) *xp = x;
	if (yp) *yp = y;
	if (wp) *wp = w;
	if (hp) *hp = h;
}

// Compute the length of each segment,
// and the total length.
// This should be called at initialization,
// and when the outline is scaled.
//
void Outline::invalidateMeasurements()
{
	//cerr << "! Enter Outline::invalidateMeasurements()\n";
	//SHOW(size());
	totalLength = -1;
	for (unsigned int i=0; i<size(); i++) {
	    info[i].invalidateMeasurements();
	}
	//cerr << "! exit Outline::invalidateMeasurements()\n";
}

double Outline::getTotalLength() const {
	if (totalLength < 0) {
		totalLength = 0;
		for (unsigned i=0; i<size(); i++) {
			totalLength += info[i].getDist();
		}
	}
	return 	totalLength;	
}

// Get the rectangular bounds of the outline after translation
//
void Outline::rect(double *xp, double *yp, double *wp, double *hp) const
{
	if (xp) *xp = translateX + (x * scaleX);
	if (yp) *yp = translateY + (y * scaleY);
	if (wp) *wp = scaleX * w;
	if (hp) *hp = scaleY * h;
}

void Outline::toRect(double newx, double newy, double neww, double newh,
            bool aspect = false, double xAlign = 0, double yAlign = 0)
{
	char *err = 0;
	if (w == 0)
	 err = "Outline::toRect: attempt to scale an outline with zero width";
	else if (h == 0)
	 err = "Outline::toRect: attempt to scale an outline with zero height";
    if (err) {
    	SHOW(err);
    	throw err;
    }

	if (aspect == false)
	{
 		scaleX = neww / w;
 		scaleY = newh / h;
 		translateX = newx - (x * scaleX);
 		translateY = newy - (y * scaleY);
	}
	else
	{
		scaleX = neww / w;
		scaleY = newh / h;
		if (scaleX > scaleY) {
			yAlign = 0;
			if      (xAlign < 0) xAlign = 0;
			else if (xAlign > 1) xAlign = 1;
			scaleX = scaleY;
		} else {
			xAlign = 0;
			if      (yAlign < 0) yAlign = 0;
			else if (yAlign > 1) yAlign = 1;
			scaleY = scaleX;
		}
		translateX = newx - (x * scaleX) + xAlign * (neww - (w * scaleX));
		translateY = newy - (y * scaleY) + yAlign * (newh - (h * scaleY));
	}
    invalidateMeasurements();
}



