/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef OUTLINE_INCLUDED
#define OUTLINE_INCLUDED

/****************************************************************
**
**	Outline represents a convex polygon
**	enclosing a set of points.
**
****************************************************************/

#include "point.h"
#include "outline_constants.h"
#include <vector>

#undef ENABLE_SHOW
#undef SHOW_NAME
#define SHOW_NAME "Outline"
#include "show.h"

class Outline
{

// Copy constructor and assignment are not allowed.
// Data members contain pointers that would be made
// invalid by copying.
//
Outline(const Outline& ) { throw "Outline: no copy constructor\n"; }
Outline& operator=(const Outline &) { throw "Outline: no assignment\n"; }

public:

Outline();
Outline(const vector<Point> &vin);

template <class In>
Outline::Outline(In b, In e); // takes two iterators to Point

~Outline();

/* Describes a point of the outline
 */
friend class PtInfo {
	Outline * mp;
	int index;
	int type;
	mutable double angle; // angle, in radians, with midpoint of outline bounds
	mutable double dist;     // distance to next point
	friend class Outline;
	friend ostream& operator<<(ostream& os, const Outline::PtInfo& p);
	mutable bool angleValid;
	mutable bool distValid;
	void invalidateMeasurements() { angleValid = distValid = false; }
 public:
	PtInfo(Outline * m, int i, int t);
	Point getPoint() const;
	void getPoint(double *x, double *y) const {
		Point p = getPoint();
		*x = p.x();
		*y = p.y();
	}
	int getIndex() const { return index; }
	int getType() const { return type; }
	double getRadians() const;
	double getDist() const;
	const PtInfo * nextInfo() const;
	const PtInfo * prevInfo() const;
	//friend class OutlineCallback;
};


// Gets the points of the outline, transformed.
void getPoints(vector<Point> &vout) const;

// Gets the center of the rectangular bounds, transformed.
Point getCenter() const;
void  getCenter(double *, double *) const;

// Gets the points of the outline, untransformed.
// The caller should not modify the contents.
const vector<Point> & getVector() const;

// Gets the number of points
unsigned int size() const { return points.size(); }

// Gets the rectangular bounds of the untransformed outline
void getVectorBounds(double *x, double *y, double *w, double *h) const;

// Gets the points using a callback function or object.
// The callback must take a Point (by value) and an int (the category
// of the point, as enumerated in outline_parts.h).
template <class Callback> void getPointsByCallback(Callback &callback) const;
template <class Callback> void getPointsByCallback(Callback *callback) const;

#define OUTLINE_OCT_ENUM
#ifndef OUTLINE_OCT_ENUM
static const int octMINX = 0;
static const int octLOHI = 1;
static const int octMAXY = 2;
static const int octHIHI = 3;
static const int octMAXX = 4;
static const int octHILO = 5;
static const int octMINY = 6;
static const int octLOLO = 7;
static const int octNONE = -1;
#else
enum OctNumber {
octMINX = 0,
octLOHI = 1,
octMAXY = 2,
octHIHI = 3,
octMAXX = 4,
octHILO = 5,
octMINY = 6,
octLOLO = 7,
octNONE = -1};

#endif

static bool isCornerOct(OctNumber oct) { return oct & 1; }

// Iterates through a section of the outline, stopping
// when the given predicate returns true.
// Returns a pointer to the PtInfo object.
template <class Pred>
const Outline::PtInfo *
Outline::findPoint(int oct, Pred & pred) const
{
	//cerr << "enter findPoint\n";
	checkOct(oct);
	if (size() == 0)
		return 0;
		
	// When size() > 0,
	// every section of the outline has at least one point.
	//
	int si, ei;
	if (oct == octNONE) {
		si = 0;
		ei = size() - 1;
	}
	else {
		si = octStart[oct];
		ei = octEnd[oct];
	}
	//SHOW(int(oct));
	//SHOW2(si,ei);
	if (si == -1 || ei == -1) {
		//cerr << "findPoint: no points for oct " << int(oct)
		//     << ", returning 0" << endl;
		return 0;
	}
	
	// ei == si if exactly one point
	for (int i=si ;;) {
		const PtInfo * p = infoPtr(i);
		if (pred(p)) {
			//cerr << "findPoint: found it, *p=" << *p << endl;
			return p;
		}
		if (i == ei) break;
		i = (i+1) % size();
	}
	//cerr << "findPoint: not found, returning 0" << endl;
	return 0;
}

// Transforms the outline so that it fits into a given rectangle
// If aspect is true, the height-to-width ratio is maintained, and
// xAlign or yAlign is used to align the figure within the rectangle
// along the dimension that doesn't fit precisely.
//
// Note: this will bomb if the untransformed outline's
// rectangular bounds (before this operation)
// has zero width or zero height.
//
void toRect(double x, double y, double w, double h,
            bool aspect = false, double xAlign = 0, double yAlign = 0);

void rect(double *x, double *y, double *w, double *h) const;

// Gets a point of the outline, transformed
Point getPoint(int i) const;
void  getPoint(int i, double *oy, double *oy) const;

// Transforms a point using the current transformation
Point transformPoint(double ix, double iy) const
{
	return Point(ix * scaleX + translateX,
		         iy * scaleY + translateY);
}

void  transformPoint(double ix,
                     double iy,
                     double *ox,
                     double *oy) const
{
	*ox = ix * scaleX + translateX;
	*oy = iy * scaleY + translateY;
}

Point transformPoint(Point p) const
{
	return transformPoint(p.x(), p.y());
}

//static OctNumber degreesToOctant(double d);

static double canonicalNumber(double d, double lim);
static double canonicalDegrees(double d);
static double canonicalRadians(double d);
static ostream& sendType(ostream& os, int t); // display PtInfo::type

double getTotalLength() const;

private:

const Outline::PtInfo * infoPtr(int i) const { return &info[i]; }

friend class OutlineCallback;

vector<PtInfo> info;     //
vector<Point> points;  // points of the outline
//PtInfo * info;
//Point *points;
//int npoints;

// indices into info and points
int octStart[8];
int octEnd[8];
void setOctStartIndex(int oct, int index);
void setOctEndIndex  (int oct, int index);

static bool badOct(int a) { return a<0 || a>7; }
static void checkOct(int o) { if (badOct(o)) throw("bad octant number\n"); }



OctNumber octNumber(int plaintype) {
	
	switch(plaintype) {
	case OutlineConstants::MINX : return octMINX;
	case OutlineConstants::LOHI : return octLOHI;
	case OutlineConstants::MAXY : return octMAXY;
	case OutlineConstants::HIHI : return octHIHI;
	case OutlineConstants::MAXX : return octMAXX;
	case OutlineConstants::HILO : return octHILO;
	case OutlineConstants::MINY : return octMINY;
	case OutlineConstants::LOLO : return octLOLO;
	default:
		cerr << "expected a plain type, got " << plaintype << endl;
		throw(-1);
	}
    return octNONE; // never happens
}

// the rectangular bounds
double x;
double y;
double w;
double h;

//
double translateX;
double translateY;
double scaleX;
double scaleY;

mutable double totalLength;
void invalidateMeasurements();

void init(const vector<Point> &);
template <class In> void init(In b, In e);
};

ostream& operator<<(ostream& os, const Outline::PtInfo& p);

#ifdef OUTLINE_OCT_ENUM
inline ostream& operator<<(ostream& os, const Outline::OctNumber& e) {
	os << static_cast<int>(e);
	return os;
}
#endif

#endif
