/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 *	Implementation details for Outline.
 *	Code which uses Outline need NOT include this file.
 */

#include <cmath>
#include "point.h"
#include "outline.h"
#include "outline_callback.h"

#undef ENABLE_SHOW
#undef SHOW_NAME
#define SHOW_NAME "OutlineCallback"
#include "show.h"

// This method is invoked while traversing the
// outline.  It saves the points and information
// about them.
//
void OutlineCallback::operator()(Point p, int type) {
	
		//SHOW2(p,type);
		
		Outline::PtInfo info(m, index, type);
		//SHOW(info.getIndex());
		//SHOW(info); // can't show until point added to outline.
		m->points.push_back(p);
		m->info.push_back(info);
		//SHOW(info);
		
		// Remember the beginning of each section (octant).
		// This assumes a knowledge of the order of traversal.
		//
		// For the first point, start at octant 0.
		// For all others, start at the most "forward" octant
		// to which the previous point belongs.
		//
		//SHOW(currOct);
		while (currOct < 8) {
			if (checkType(type, currOct)) {
				if (currOct < 7)
					currOct = checkForward(type, currOct + 1);
				break;
			} else {
				currOct++;
			}
		}
		
		index++;  // next index into m->points and m->info
}

// Determine if the given type
// should be placed in the given octant.
// If so, save the value of the member variable `index'
// as the last index seen for this octant.
// Also, save the first index seen for
// this octant.
//

bool OutlineCallback::simpleCheck(int type, int oct) {
	
	//Outline::checkOct(oct); // abort if bad param
    if (Outline::badOct(oct)) {
    	SHOW2(index, type);
    	SHOW(oct);
    	throw "simpleCheck: bad oct";
    }
	
	if (type & OutlineConstants::getMask(oct)) {
		last[oct] = index;
		if (!startIsSet[oct]) { // first occurrence of this type
			startIsSet[oct] = true;
			m->setOctStartIndex(oct, index);
		}
		return true;
	}
	return false;
}

// Determine if the given type
// should be placed in the given octant.
// If so, see if it should also be in any "previous"
// octant in the traversal order.
//
// For example, if the type is that of a point
// on the minimum-x side of the outline, it should be
// placed in the MINX group, and the point
// might be at the (min-x max-y) corner,
// in which case it should be placed in the LOHI group
// and the MAXY group, and possibly the HIHI and MAXX groups
// if the outline is one-dimensional, and all groups
// if the outline is a single point.
//
// Calls simpleCheck (which see).
//
bool OutlineCallback::checkType(int type, int oct) {
	
	bool isFirst = !startIsSet[oct]; // if not set before simpleCheck
	bool isType = simpleCheck(type, oct);
	if (isFirst && isType) {
		checkType(type, prevOct(oct));			
	}
	return isType;
}

// Determine if the given type
// should be placed in the given octant.
// If so, see if it should also be in any "following"
// octant in the traversal order.
// Calls simpleCheck (which see).
//
int OutlineCallback::checkForward(int type, int oct) {

    //Outline::checkOct(oct); // abort if bad param
    if (Outline::badOct(oct)) {
    	SHOW2(index, type);
    	SHOW(oct);
    	throw "checkForward: bad oct";
    }

	int newIndex = oct - 1;
	for (int i=oct ; i < 8; i++) {
		if (simpleCheck(type, i)) {
			newIndex = i;
		} else {
			break;
		}
	}
	return newIndex;
}

void OutlineCallback::finish() {
	for (int i=0; i<8; i++) {
		m->setOctEndIndex(i, last[i]);
	}
}

			
