/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef LABEL_OUTLINED_POLYGON_H
#define LABEL_OUTLINED_POLYGON_H

/****************************************************************
**
** OutlinedPolygon is a type of QCanvasPolygon,
** which has an Outline.
**
** Subclasses should call setOutline(); this will
** compute the "area points" (a Qt concept) and
** call QCanvasPolygon::setPoints().
** The subclass may draw on the outline, but normally should
** not draw on the QCanvasPolygon::areaPoints()
** 
**
****************************************************************/

#include <iostream>
#include <vector>

#include <qcanvas.h>
#include "outline.h"

// macros for debugging
#undef ENABLE_SHOW
#undef SHOW_NAME
#define SHOW_NAME "OutlinedPolygon"
#include "show.h"

/*
 *	OutlinedPolygon is a QCanvasPolygon
 *	that has an Outline.
 */
class OutlinedPolygon : public QCanvasPolygon
{
public:
	OutlinedPolygon(QCanvas *cp)
		: QCanvasPolygon(cp)
	{
	}
		
	~OutlinedPolygon() { delete outline; }
	
	Outline *getOutline() const { return outline; }
	
	Point getCenter() const {
		Point c = outline->getCenter() + Point(x(),y());
#if 0
		SHOWF("enter ---\n");
		SHOW2( outline->getCenter(), Point(x(),y()) );
		QRect r = boundingRect();
		SHOW2(r.x(), r.y()); SHOW2(r.width(), r.height());
		SHOW0 << "| " <<
			 SHOW_NMVAL(r.center().x()) << ", " <<
			 SHOW_NMVAL(r.center().y()) << endl;
		SHOW0 << "| " << SHOW_NMVAL(c) << endl;
		double ox, oy, ow, oh;
		outline->rect(&ox, &oy, &ow, &oh);
		SHOW2(ox, oy); SHOW2(ow, oh);
		SHOWF("exit ---\n");
#endif
		return c;
	}
	
protected:

	// Sets the outline for this object.
	// Computes the "outer path" (which is
	// the outline converted to QPoints).
	// Computes the "area points" (a Qt concept)
	// to be slightly larger than the outer path,
	// and calls QCanvasPolygon::setPoints().
	// The object may draw on the outer path,
	// but should not draw on the lines defined
	// by areaPoints().
	//
	void setOutline(Outline *m) {
	
		outline = m;
    	outer_path = QPointArray(m->size());
    	QPointArray area_points(m->size());
    	double cx, cy;
    	m->getCenter(&cx, &cy);
    	
    	for (unsigned int i=0; i<m->size(); i++) {
    		double px, py;
    		m->getPoint(i, &px, &py);
    		outer_path[i] = QPoint( qRound(px), qRound(py) );
    		if (px > cx) px++;
    		else if (px < cx) px--;
    		if (py > cy) py++;
    		else if (py < cy) py--;
    		
    		area_points[i] = QPoint( qRound(px), qRound(py) );
    	}
    	
    	// Work around a Qt restriction, that the polygon's
    	// drawShape() might not be called if only 2 points
    	int sz = area_points.size();
    	if (sz == 2) {
    		for (unsigned int i=0; i<2; i++) {
    			// append two points
    			area_points.putPoints( sz,2,2,
    				outer_path[1].x(), outer_path[1].y(),
    				outer_path[0].x(), outer_path[0].y() );
    		}	
    	}
    	
    	setPoints(area_points);
	}
	
	// Gets the outer path, which forms a convex polygon
	// and defines the area upon which the object may draw.
	// caller should detach() the QPointArray before
	// making changes to it.  These points should be
	// translated to the object's x() and y() location
	// before drawing.
	QPointArray getOuterPath() { return outer_path; }
	
private:
	// Don't allow copy constructor or assignment.
	// This makes easier the management of pointers
	// that are deleted by the destructor.  I.e. the destructor
	// can simply delete them because they are not shared.
	//
	OutlinedPolygon(const OutlinedPolygon &r) : QCanvasPolygon(r.canvas())
		{ throw "OutlinedPolygon: no copy constructor\n"; }
	OutlinedPolygon& operator=(const OutlinedPolygon &)
		{ throw "OutlinedPolygon: no assignment operator\n"; }
	
	Outline * outline;
    QPointArray outer_path; // an outline that can be drawn
};

#endif LABEL_OUTLINED_POLYGON_H
