/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/****************************************************************
**
**  Implementation of the Polygon class
**
****************************************************************/

#include <iostream>
#include <vector>

#include "polygon.h"
#include "rand.h"
#include "outline.h"
#include "point.h"

#include <qcanvas.h>
#include <qpainter.h>

#undef ENABLE_SHOW
#undef SHOW_NAME
#define SHOW_NAME "Polygon"
#include "show.h"

Polygon::Polygon(QCanvas *cp,
                 const vector<Point> & v,
                 const QRect & bounds)
	: OutlinedPolygon(cp)
{
	const QRect & b = bounds;
	
	// Produce an outline of the points of v,
	// tranformed to fit into bounds b.
	Outline *op = new Outline(v);
	op->toRect(b.x(), b.y(), b.width(), b.height());
	
	// transform points from v to inner_path
	int nPoints = v.size();
	inner_path = QPointArray(nPoints);
	for (int i=0; i<nPoints; i++) {
		Point p = op->transformPoint( v[i].x(), v[i].y() );
		int x = qRound( p.x() );
		int y = qRound( p.y() );
#if DEBUG_POLYGON
		if (x==b.x() && y==b.y()) {
			cerr << "---------------\n";
			cerr << "Polygon: \n";
			cerr << " "; SHOW2(i, nPoints);
			cerr << " "; SHOW2(v[i].x(), v[i].y());
			cerr << " "; SHOW2(x,y);
			cerr << " "; SHOW2(b.x(), b.y());
			cerr << " "; SHOW2(b.width(), b.height());
	    	cerr << "---------------\n";
		}
#endif
		inner_path[i] = QPoint(x, y);
	}
	
	setOutline(op);
	setDrawnFilled(true);
	setOutlineDrawn(false);
}

Polygon::~Polygon()
{
}

void Polygon::drawShape(QPainter & p)
{
	//cerr << "--enter Polygon::drawShape\n";
	p.save();
	p.translate(x(),y());
	
	if (outlineDrawn()) {
		SHOWF("drawing outline using current pen\n");
		// Draw the outline, assuming setPen() has been called.
		// See the base class, OutlinedPolygon.
		QPointArray outer_path = getOuterPath();
		int osize = outer_path.size();
		for (int i=0; i<osize; i++) {
			QPoint a = outer_path[i];
			QPoint b = outer_path[(i+1) % osize];
			p.drawLine(a.x(), a.y(), b.x(), b.y());
		}
	}
	
	// Draw the internal path, using the current brush()
#if 0
	if (inner_path.size() == 2) {
		p.setPen(brush().color());
		QPoint a = inner_path[0];
		QPoint b = inner_path[1];
		p.drawLine(a.x(), a.y(), b.x(), b.y());	
	}
	else
#endif
	if (drawnFilled()) {
		p.drawPolygon(inner_path);
	}
	else {
		p.setPen(brush().color());
		int psize = inner_path.size();
		for (int i=0; i<psize; i++) {
			QPoint a = inner_path[i];
			QPoint b = inner_path[(i+1) % psize];
			p.drawLine(a.x(), a.y(), b.x(), b.y());
		}
	}
	
	p.restore();
	
	//cerr << "--exit Polygon::drawShape\n";
	
}

