/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef INCLUDED_OUTLINE_CORNER
#define INCLUDED_OUTLINE_CORNER

#include <string>
#include <vector>
#include <functional>
#include <cassert>
#include <iostream>

#undef SHOW
#define SHOW(X) cerr << #X "=" << (X) << endl
#undef SHOW2
#define SHOW2(X,Y) cerr << #X "=" << (X) << ", " #Y "=" << (Y) << endl
#undef SHOW2P
#define SHOW2P(X,Y) cerr << #X "=" << (X).x() << "," << (X).y() << "; " \
	                        #Y "=" << (Y).x() << "," << (Y).y() << endl


struct SlopeException {};

/*
 *	The Corner class is used by OutlineScanner to
 *	calculate the path around one corner of the
 *	rectangular bounds of a polygon.  This class is not
 *	intended to be used directly, but only by the
 *	OutlineScanner template.
 *
 *	Template parameters:
 *	P = the class of points, which must have x() and y() methods which
 *		yield the coordinates of the point.
 *	    P should have value semantics.
 *	Num = the numeric type of the x() and y() methods of P.
 *      This should be a floating point type.
 *	Cmp1 = a binary comparator which determines whether a point will be
 *		included in the outline, based on its y value.  For example,
 *		if Cmp1 is less<Num>, a point will be included only if it's "below"
 *		the current gap in the corner.  Cmp1 must have a default constructor.
 */
template <class P, class Num, class Cmp1>
class Corner {

	typedef long double slope_t;
	string name;  // for debug
	
	// if valid is true, the slope and y-intercept have meaningful values
	bool valid;
	slope_t slope;
	slope_t y_intercept;
	P p1, p2;
	
	// the comparators are used to determine which points
	// are included.
	// They cannot be changed.
	Cmp1 yCmp;
	greater<slope_t> slopeCmp;
	
	// "first" is the first point of those remaining
	// in the path for this corner.  See update() and traverseWithCallback().
	// The value of first is valid iff haveFirst is true.
	bool haveFirst; //
	P first;
	slope_t firstSlope;
	
	// A container for points under consideration.
	vector<P> q;
	
	ostream & say() {
		// note: Qt's QPoint type can't be sent to an output stream
		// so we have to explicitly send the x() and y() values.
		//cerr << "corner " << name << "[" << p1 << "," << p2 << "]: ";
		cerr << "corner " << name << "[" <<
			 p1.x() << "," << p1.y() << ":" <<
			 p2.x() << "," << p2.y() << "]: ";
		return cerr;
	}
	
public:

	Corner() : name("--"), valid(false), p1(0,0), p2(0,0)  {
		//say() << "constructor\n";
	}
	~Corner() {
		//say() << "destructor\n";
	}
	

	// Gets the slope and y-intercept of a line,
	// given the x and y values of two points.
	// The order of the points does not affect the results.
	//
	// the number parameters must be "by value", since swap() is used.
	static void getSlopeIntercept(Num x1, Num y1, Num x2, Num y2,
	                              slope_t *s=0, slope_t *i=0)
	{
		if (x1 < x2) {
			swap(x1,x2); // parameters called be value, not reference
			swap(y1,y2);
		}
		if (x1==x2) { cerr <<"SlopeException\n"; throw SlopeException(); }
		if (s) *s = static_cast<slope_t>(y2-y1)/(x2-x1);
		if (i) *i = (y1 - (*s) * x1);
	}
	static void getSlopeIntercept(const P &a, const P &b,
	                                 slope_t *s=0, slope_t *i=0)
	{
		getSlopeIntercept(a.x(),a.y(),b.x(),b.y(), s, i);
	}
	static slope_t getSlope(const P &a, const P &b) {
		slope_t s;
		getSlopeIntercept(a, b, &s);
		return s;
	}
	
	// Returns the y value for a point on the line with a given x.
	// (potential rounding errors here).
	//
	static Num getY(Num x, slope_t s, slope_t intercept) {
		return static_cast<Num> (x * s + intercept);
	}
	
	// Initializes the Corner object,
	// using a name (for debugging) and two points which cross the corner.
	// The two points should be given in traversal order.
	//
	void init(const string & nm,
	          const Num &x1, const Num & y1,
	          const Num &x2, const Num & y2)
	{
		name = nm;
		//assert(x1<=x2);
		p1 = P(x1,y1);
		p2 = P(x2,y2);
		//Num xdiff = (x2-x1);
		valid = (x1 != x2) && (y1 != y2);
		if (!valid)
			return;
		// denominator always non-zero
		try { getSlopeIntercept(x1,y1,x2,y2, &slope, &y_intercept); }
		catch (SlopeException) { say(); SHOW2(x1,y1); SHOW2(x2,y2); throw; }
		
		haveFirst = false;
		q.clear();
		assert(slope != 0); // in case of error in getSlopeIntercept
	}
	
	
	bool isValid() { return valid; } // call after init()
	
	// Considers a point for inclusion in the corner's outline.
	// Should be called after init and before traverse.
	void update(const P & p) {
		if (!valid) return;
		//say() << "enter update\n";
		
		// The endpoints p1 and p2 of the line segment should be ignored.
		// The next line does so the easy way, preventing some
		// unresolved rounding errors.
		if (p == p1 || p == p2) return;
		
		Num x = p.x();
		Num y = p.y();
		Num yline = getY(x, slope, y_intercept);
		//say() << "update: checking point " << p << ", ";
		//SHOW2(y, yline);
		//SHOW2(slope, x * slope + y_intercept);
		
		// Consider only points above or below the p1,p2 segment.
		// first is set to the "greatest" one;
		// all others are put in list q.
		if (yCmp(y, yline)) {
			//say() << "update: adding point " << p << endl;
			//SHOW2(y,yline); SHOW2P(p1,p2); SHOW2(slope,y_intercept);
			slope_t s;
			try { s = getSlope(p1,p); } // don't expect infinite slope
			catch (SlopeException) {
			 	say(); SHOW2P(p1,p); SHOW2(y,yline); SHOW2(slope,y_intercept);
			 	throw;
			}
			if (!haveFirst) {
				first = p;
				firstSlope = s;
				haveFirst = true;
			} else {
				if (slopeCmp(s,firstSlope)) {
					q.push_back(first);
					first = p;
					firstSlope = s;
				}
				else q.push_back(p);
			}
		}
		//say() << "exiting update\n";
	}
	
	// Gets the points of the corner.
	// Should be called only once, after init() and update().
	// This method is invoked by Corner.
	//
	// Args:
	// 	cb = a ptr to function or object, which takes P and int.
	//  type = a value passed to *cb.
	//
	template <class Callback>
	void traverseWithCallback(Callback *cb, int type) {
		//say() << " traverse: "; SHOW(valid);
		if (!valid) return;
		//say() << " traverse: "; SHOW2(haveFirst, q.size());
		vector<P> q2;
		while (haveFirst) {
			P temp = first; // prevent change to first
			(*cb)( temp, type );
			//say() << "  traversed " << first << endl;
			P curr = first;
			slope_t curr_slope;
			slope_t curr_intercept;
			// don't expect infinite slope
			try { getSlopeIntercept(curr, p2, &curr_slope, &curr_intercept); }
			catch (SlopeException) { say(); SHOW2P(curr,p2); throw; }
			haveFirst = false;
			vector<P>::const_iterator i;
			
			// find the "greatest" point, put others in q2
			for (i=q.begin(); i!=q.end(); i++) {
				
				Num x = i->x(), y = i->y();
				Num yline = getY(x, curr_slope, curr_intercept);
				
				// ignore the point if y-value not within current line
				if (yCmp(y, yline)) {
					slope_t s;
					try { s = getSlope(curr, *i); } // don't expect infinite slope
					catch (SlopeException) { say(); SHOW2P(curr, *i); throw; }
					if (!haveFirst) {
						haveFirst = true;
						first = *i;
						firstSlope = s;
					} else {
						if (slopeCmp(s,firstSlope)) {
							q2.push_back(first);
							first = *i;
							firstSlope = s;
						}
						else {
							q2.push_back(*i);
						}
					}
				}
				else {
					//say() << "  skipping  " << *i << endl;
				}
			}
			q.clear();
			q.swap(q2);
		} // end while haveFirst
	} // end method

	
};


#endif

