/* Copyright 2002 Scott Haywood

This file is part of the Label program.

Label is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Label is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Label; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef INCLUDED_OUTLINE_SCANNER
#define INCLUDED_OUTLINE_SCANNER

#include <string>
#include <vector>
#include <functional>
#include <cassert>
#include "outline_corner.h"
#include "outline_constants.h"

using namespace OutlineConstants;

#undef SHOW
#define SHOW(X) cerr << #X "=" << (X) << endl
#undef SHOW2
#define SHOW2(X,Y) cerr << #X "=" << (X) << ", " #Y "=" << (Y) << endl

/*
 * The OutlineScanner template computes a convex polygon
 * enclosing a sequence of points in 2-dimensional space.
 *
 * Template parameters:
 * In = an input iterator which yields a sequence of objects of type P.
 * P = a class of "points", which must have methods x() and y()
 *		that return the coordinates of the point.
 *		It must have a constructor which takes two items of type Num.
 * Num = the numeric type of the x() and y() methods of P.
 *		This should be a floating-point type.
 *
 * Example:  Create random points in v.  Create the outline of v in v2.
 *	vector<Point> v;
 *	for(int i=0;i<100;i++) v.push_back(Point(rand(),rand()));
 *	OutlineScanner<vector<Point>::iterator,Point> o1(v.begin(), v.end());
 *	vector<Point> v2;
 *	o1.traverse(back_inserter(v2));
 */

/*
Implementation:

Let S the set of points for which the outline is desired.
Find the "bounding rectangle" of S, which is described by
the mininum and maximum values of the x and y coordinates.

While finding the bounding rectangle, find the extreme points on
each side of the rectangle.  For example, for all the points having the
minimum x value, find the ones with the least and greatest y values.
This results in at most two points per side of the bounding rectangle;
these points are called the "octangonal limits" for want of a better name.

Use the Corner template (defined elsewhere) to connect the octagonal limits
of one side to the octagonal limits of the next adjacent side.

The desired result, the convex polygon bounding S, is obtained by properly
ordering the octangonal limits with the points determined by the Corner objects.

*/
template <class In, class P, class Num=double>
class OutlineScanner
{
    //typedef OutlineScanner<In, P, Num> Outline_type;

	In i1, i2;

	// values defining the rectangular bounds
	Num minx;
	Num miny;
	Num maxx;
	Num maxy;
	
	// min and max values on the rectangular bounds.
	Num octlim[8];
	
	// the minimum-x side of the rectangular bounds
	Num * xlo;
	Num & xlo_miny;
    Num & xlo_maxy;
    //  the maximum-x side
    Num * xhi;
	Num & xhi_miny;
    Num & xhi_maxy;
    // the minimum-y side
    Num * ylo;
	Num & ylo_minx;
    Num & ylo_maxx;
    // the maximum-y side
    Num * yhi;
	Num & yhi_minx;
    Num & yhi_maxx;

    // Corners
    Corner<P,Num, less<Num> >    corner_lolo;
    Corner<P,Num, greater<Num> > corner_lohi;
    Corner<P,Num, less<Num> >    corner_hilo;
    Corner<P,Num, greater<Num> > corner_hihi;


    template <class N, class Cmp>
    void update_side(N a, N b, N & lim, N * side, Cmp cmp) {
 		if (cmp(a, lim)) {
    		lim = a; side[0] = side[1] = b;
    	}
    	else if (a == lim) {
    		if (b < side[0]) side[0] = b;
    		if (b > side[1]) side[1] = b;
    	}
    }

    // Computes the rectangular bounds, and the min and max
    // points on each side of the bounds.
    //
    void initSides() {
    	if (i1 == i2) return;
    	minx = maxx = ylo_minx = ylo_maxx = yhi_minx = yhi_maxx = i1->x();
        miny = maxy = xlo_miny = xlo_maxy = xhi_miny = xhi_maxy = i1->y();
        In i = i1;
        for (i++; i != i2; i++) {
        	Num x = i->x();
        	Num y = i->y();
        	update_side(x, y, minx, xlo, less<Num>());
        	update_side(x, y, maxx, xhi, greater<Num>());
        	update_side(y, x, miny, ylo, less<Num>());
        	update_side(y, x, maxy, yhi, greater<Num>());
        }
     }

     // Initializes the Corner objects, using
     // the results of initSides().
     //
     void initCorners() {
     	// points given to init() are in traversal order
     	corner_lolo.init("lolo",ylo_minx,miny,minx,xlo_miny);
		corner_lohi.init("lohi",minx,xlo_maxy,yhi_minx,maxy);
		corner_hihi.init("hihi",yhi_maxx,maxy,maxx,xhi_maxy);
		corner_hilo.init("hilo",maxx,xhi_miny,ylo_maxx,miny);
		for (In i = i1; i != i2; i++) {
			 corner_lolo.update(*i);
			 corner_lohi.update(*i);
			 corner_hihi.update(*i);
			 corner_hilo.update(*i);
		}
	}


public:

    OutlineScanner(In first, In last) :
    	i1(first),
    	i2(last),
    	
    	xlo(&octlim[0]),
    	xlo_miny(octlim[0]),
    	xlo_maxy(octlim[1]),
    	
    	xhi(&octlim[4]),
    	xhi_miny(octlim[4]),
    	xhi_maxy(octlim[5]),
    	
    	ylo(&octlim[6]),
    	ylo_minx(octlim[6]),
    	ylo_maxx(octlim[7]),
    	
    	yhi(&octlim[2]),
    	yhi_minx(octlim[2]),
    	yhi_maxx(octlim[3])
    	
    {
    	minx = maxx = miny = maxy = 0;
    	for (int i=0; i<8; i++) { octlim[i] = 0; }
    	initSides();
    	initCorners();
    }


    Num getMinx() { return minx; }
    Num getMaxx() { return maxx; }
    Num getMiny() { return miny; }
    Num getMaxy() { return maxy; }
    	
    friend ostream & operator<< <> (ostream &os, const OutlineScanner<In,P,Num> &o);

    /*
     * Get the octagonal limits of the path,
     * i.e. the max and min points on each side of the rectangular bounds.
     * There will not necessarily be 8 unique points;
     * for example, if an elongated diamond shape is tilted
     * so that its acute points have more extreme x and y values
     * than the other points, there will be only two.
     * Some points of the path can be outside the "octagon";
     * for example, a nearly-circular
     * equilateral polygon would have most of its points outside.
     * The output points are ordered to form a path,
     * starting at the minimum-x side, then the maximum-y side.
     */
    void getOctagonalLimits(P (&pa)[8]) const {
    	//cout << "getOctangonalLimits: xlo_miny=" << xlo_miny <<endl;
    	pa[0] = P(minx, xlo_miny);
    	pa[1] = P(minx, xlo_maxy);
    	pa[2] = P(yhi_minx, maxy);
    	pa[3] = P(yhi_maxx, maxy);
    	pa[4] = P(maxx, xhi_maxy);
    	pa[5] = P(maxx, xhi_miny);
    	pa[6] = P(ylo_maxx, miny);
    	pa[7] = P(ylo_minx, miny);
	}
	
#if 0	
	// like above, but get the points using an output iterator.
	template <class Out>
	void getOctagonalLimits(Out out) {
		P p[8];
		getOctagonalLimits(p);
		for (int i=0; i<8; i++) {
			*out++ = p[i];
		}
	}
#endif


private:

	// Gets the type of a point which is NOT on a side.	
	int getCornerType(const P & p) {
		Num x = p.x();
		Num y = p.y();
		if (x < yhi_minx && y > xlo_maxy) return LOHI;
		if (x > yhi_maxx && y > xhi_maxy) return HIHI;
		if (x > ylo_maxx && y < xhi_miny) return HILO;
		if (x < ylo_minx && y < xlo_miny) return LOLO;
		return 0;
	}
	
	/*
	*	Gets the type of a point which is on at least
	*	one of the sides of the rectangular bounds.
	*	See OutlineConstants
	*/
	int getSideType(const P & p) {
		int t = 0;
		Num x = p.x();
		Num y = p.y();
		if (x == minx) t |= MINX;
		if (y == miny) {
			t |= MINY;
			if (t & MINX) t |= LOLO;
		}
		if (x == maxx) {
			t |= MAXX;
			if (t & MINY) t |= HILO;
		}
		if (y == maxy) {
			t |= MAXY;
			if (t & MINX) t |= LOHI;
			if (t & MAXX) t |= HIHI;
		}
		return t;
	}
public:	
	int getType(const P & p) {
		int t = getSideType(p);
		if (t) return t;
		else return getCornerType(p);
	}

private:	
	template <class Callback>
	void deliverSidePoint(P & p, Callback *cb)
	{
		(*cb)(p, getSideType(p));
	}	
public:	
	// Traverse the points of the outline,
	// passing them to a callback function (or object)
	//
	template <class Callback>
	void traverseWithCallback(Callback *cb) // by reference, below
	{
		P p[8];
		getOctagonalLimits(p);
		// p[] starts with (minx,xlo_miny) and ends with (ylo_minx,miny)
		
		// lowest point on the minx side
		deliverSidePoint( p[0], cb );
		
		// highest point on the minx side
		if (xlo_miny != xlo_maxy)
			deliverSidePoint( p[1], cb );
		
		//cout << "traverse: calling corner_lohi.traverseWithCallback\n";
		corner_lohi.traverseWithCallback(cb, LOHI);
		
		// lowest point on the maxy side
		if (yhi_minx != minx)
			deliverSidePoint( p[2], cb );
		
		// highest point on the maxy side
		if (yhi_minx != yhi_maxx)
			deliverSidePoint( p[3], cb );
		
		//cout << "traverse: calling corner_hihi.traverseWithCallback\n";
		corner_hihi.traverseWithCallback(cb, HIHI);
		
		// highest point on the maxx side
		if (xhi_maxy != maxy)
			deliverSidePoint( p[4], cb );
		
		// lowest point on the maxx side
		if (xhi_maxy != xhi_miny)
			deliverSidePoint( p[5], cb );
		
		//cout << "traverse: calling corner_hilo.traverseWithCallback\n";
		corner_hilo.traverseWithCallback(cb, HILO);
		
		// highest point on the miny side
		if (ylo_maxx != maxx)
			deliverSidePoint( p[6], cb );
			
		// lowest point on the miny side
		if (ylo_maxx != ylo_minx && ylo_minx != minx)
			deliverSidePoint( p[7], cb );
		
		//cout << "traverse: calling corner_lolo.traverseWithCallback\n";
		corner_lolo.traverseWithCallback(cb, LOLO);
	}

	// Same as above, but callback given as a reference
	//
	template <class Callback>
	void traverseWithCallback(Callback &cb)
	{
		traverseWithCallback(&cb);
	}
		
	
	template <class Out>
	class Callback_link {
		Out out;
	public:
		Callback_link(Out o) : out(o) {}
		void operator()(P &p, int) { *out++ = p; }
	};
	
	// Gets the points which form the convex polygon.
	//
	template <class Out>
	void traverse(Out out) // don't pass by reference
	{
		Callback_link<Out> cl(out);
		traverseWithCallback(&cl);
	}
	
	// Get a vector containing the outline points.
	// Note that this does the traverse()
	// each time it's called.
	vector<P> getVector() {
		vector<P> v;
		traverse(back_inserter(v));
		return v;
	}
	
    	
};

// Send an OutlineScanner object to an output stream.
// This function is used for testing and debugging.
// For example, assuming o1 is an OutlineScanner:
//	#include <iostream>
//	cout << o1;
//
template <class In, class P, class Num>
ostream & operator<<(ostream &os, const OutlineScanner<In,P,Num> &o) {
    os << "OutlineScanner{";
    os << "minx=" << o.minx << ",";
	os << "miny=" << o.miny << ",";
	os << "maxx=" << o.maxx << ",";
	os << "maxy=" << o.maxy << ",";
	os << "\n";

	os << "\"octangonal\" limits: ";
	P p[8];
	o.getOctagonalLimits(p);
	for (int i=0; i<7; i++)
		os << p[i].x() << "," << p[i].y() << ", ";
	os << p[7].x() << "," << p[7].y();
	
	os << "}";
	return os;
}



#endif
